/* RealSpeak.cpp

RealSpeak(TM) driver.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Peter Dikant <peter.dikant@mobilexag.de>
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include "RealSpeak.h"

#define TTS_OUTPUT_BUFFER 1048576 // 1 MByte output buffer

// this header will be appended to the output created by the tts system
// to create a complete wav memory structure
static unsigned char wavHeader[] = {
	0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00, // |R|I|F|F|4 byte lenght of file - 8|
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20, // |W|A|V|E|f|m|t| |
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, // |4 byte length of header|2 bytes encoding (01 = pcm)|2 bytes number of channels|
	0x40, 0x1f, 0x00, 0x00, 0x80, 0x3e, 0x00, 0x00, // |4 byte samplesrate|4 byte samplerate * bytes per sample|
	0x02, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, // |2 byte bytes per sample|2 byte bit per sample|d|a|t|a|
	0x00, 0x00, 0x00, 0x00 };

int RealSpeak::initialize(Logger*log)
{
    log->logf("<verbose> Using Realspeak as Text to Speech engine\n");
    return 1;
}

void RealSpeak::finalize(Logger*log)
{
}

/**
 * the default constructor will create the basic tts instances and also store a pointer
 * to the logging subsystem.
 * @param log a pointer to the logging object
 */
RealSpeak::RealSpeak(Logger *log, const char* engine)
    :TextToSpeech(log)
{
    log->logf("<debug> Initializing RealSpeak Engine");
    log->logf("<debug> starting with data buffer of %d bytes", TTS_OUTPUT_BUFFER);

    strcpy(enginePath, engine);
    ttsParm.nLanguage = TTS_LANG_GERMAN;
    ttsParm.nOutputType = TTS_LINEAR_16BIT;
    ttsParm.nFrequency = TTS_FREQ_8KHZ;
    ttsParm.nVoice = TTS_RS_VOICE_FEMALE;
    ttsParm.nInputDataType = TTS_DATA_TYPE_TEXT;
    ttsParm.nOutputDataType = TTS_DATA_TYPE_PCM;
    ttsParm.cbFuncs.TtsSourceCb = sourceCallback;
    ttsParm.cbFuncs.TtsDestCb = destCallback;
    ttsParm.cbFuncs.TtsEventCb = CbTtsEventNotify;
    ttsParm.cbFuncs.numCallbacks = 3;
    ttsParm.szLibLocation = enginePath;

    int ret = TtsInitialize(&hInst, NULL, &ttsParm, (void *)this);
    if (ret != 0) {
	log->logf("<error> Error initializing RealSpeak Engine: %d", ret);
    }
    log->logf("<debug> RealSpeak Engine initialized.");

    output = (unsigned char*)malloc(TTS_OUTPUT_BUFFER + 44);
    outputSize = TTS_OUTPUT_BUFFER;
    memcpy(output, wavHeader, 44);
    InitializeCriticalSection(&criticalSection);
}

TTSRETVAL RealSpeak::CbTtsEventNotify (void *pAppData, void *ppBuffer, U16 nDataSize, U16 event) {
    RealSpeak *me = (RealSpeak*)pAppData;
    //me->log->logf("<debug> Received RealSpeak Event: %d", event);
    return 0;
}

TTSRETVAL RealSpeak::sourceCallback(void *pAppData, void *data, U32 len, U32 *datasize) {
    RealSpeak *me = (RealSpeak*)pAppData;
    
    if (!me->text || !*(me->text)) {
	*datasize = 0;
	if(me->text) {
	    free(me->textstart);
	    me->text = 0;
	}
	return TTS_ENDOFDATA;
    }
    if (len >= strlen(me->text)) {
	memcpy(data, me->text, strlen(me->text));
	*datasize = strlen(me->text);
	me->text += strlen(me->text);
    } else {
	memcpy(data, me->text, len);
    	*datasize = len;
	me->text += len;
    }
    return TTS_SUCCESS;
}

void* RealSpeak::destCallback(void *pAppData, U16 nDatatype, void *data, U32 datasize, U32 *buffersize) {
    RealSpeak *me = (RealSpeak*)pAppData;
    void*ret = 0;
    //me->log->logf("<debug> TTS: DestCallback: datatype:%d data:%08x size:%d", nDatatype, data, datasize);

    while((int)(datasize+TTS_OUTPUT_BUFFER/2) > me->outputSize - me->outputPos) {
	me->outputSize += TTS_OUTPUT_BUFFER;
	void*newdata = realloc(me->output, me->outputSize + 44);
	if(!newdata) {
	    // failed, cut off
	    me->log->logf("<error> TTS: realloc(%d) failed", me->outputSize);
	    datasize = me->outputSize - me->outputPos;
	    me->outputSize -= TTS_OUTPUT_BUFFER; //revert
	    break;
	} else {
	    me->log->logf("<notice> TTS: expanded output buffer to %d", me->outputSize);
	    me->output = (unsigned char*)newdata;
	}
    }
    *buffersize = me->outputSize - me->outputPos;

    ret = (void *)&me->output[me->outputPos + 44];
    me->outputPos += (long)datasize;
    //me->log->logf("<debug> TTS: DestCallback: returning memory for %d bytes", *buffersize);
    return ret;
}

/**
 * free used ressources
 */
RealSpeak::~RealSpeak()
{
	log->logf("<debug> Deleting RealSpeak Engine");
	int ret = TtsUninitialize(hInst);
	if (ret != 0) {
	    log->logf("<error> Error closing RealSpeak Engine: %d", ret);
	}
	free(output);
	DeleteCriticalSection(&criticalSection);
}

/*
 * load a User Dictionary
 */
void RealSpeak::loadDictionary(const char* dictionary)
{
	int ret;
	if(dictionary) {
	    ret = TtsLoadUsrDict(0, &userdict, (char*)dictionary);
	    if(ret != 0) {
		log->logf("<error> Couldn't load dictionary \"%s\"", dictionary);
	    } else {
		ret = TtsEnableUsrDict(hInst, userdict);
		if(ret != 0) {
		    log->logf("<error> Couldn't enable dictionary \"%s\"", dictionary);
		} else {
		    log->logf("<notice> Dictionary \"%s\" loaded", dictionary);
		}
	    }
	}
}

/**
 * set implementation specific parameters
 */
void RealSpeak::setParams(const char *key, const char *value)
{
	/* test which parameters are for us */
	if(!strncmp(key, "realspeak.", 10))
	{
	    const char*key2 = &key[10];
	    if(!strcmp(key2, "dictionary"))
	    {
		loadDictionary(value);
	    }
	}
}

/**
 * this method will convert a given text to speech and output this over the microsoft
 * audio mapper, thus creating direct audio output on the soundcard.
 * @param text the text to be converted to speech
 */
void RealSpeak::text2Audio(const char *text)
{
	log->logf("<verbose> text2Audio is not implemented in RealSpeak");
}

/**
 * create speech directly to a wave file
 * @param text the text to be spoken
 * @param filenam name and path of the file to be created
 */
void RealSpeak::text2File(const char *text, const char *filename)
{
    EnterCriticalSection(&criticalSection);
    this->text = this->textstart = strdup(text);

    // start conversion
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputPos);

    unsigned int wavelen = outputPos + 36;
    memcpy(output + 4, &wavelen, 4);
    wavelen = outputPos;
    memcpy(output + 40, &wavelen, 4);

    FILE *fp = fopen(filename, "wb");
    fwrite(output, sizeof(unsigned char), outputPos + 44, fp);
    fclose(fp);
    LeaveCriticalSection(&criticalSection);
}


/**
 * this method will create speech that is stored in a memory buffer. the created speech has the format pcm mono with 
 * 8kHz sampling rate and 16 bit. The memory structure can be directly used for input to the capi
 * subsystem.
 * @param text this text will be spoken
 * @return oubuffer pointer to the output buffer (the memory will be allocated)
 * @return size the size of the allocated memory
 */
void RealSpeak::text2Stream(const char *text, unsigned char **outbuffer, unsigned long *size)
{
    EnterCriticalSection(&criticalSection);
    this->text = this->textstart = strdup(text);

    // start conversion
    outputPos = 0;
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	*size = 0;
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputPos);

    
    *outbuffer = (unsigned char*)malloc (outputPos + 44);
    unsigned int wavelen = outputPos + 36;
    memcpy(&output[4], &wavelen, 4);
    wavelen = outputPos;
    memcpy(&output[40], &wavelen, 4);
    memcpy(*outbuffer, output, outputPos + 44);
    *size = outputPos + 44;
    
/*    FILE *fp = fopen("temp.wav", "wb");
    fwrite(*outbuffer, sizeof(unsigned char), outputPos + 44, fp);
    fclose(fp); */
    LeaveCriticalSection(&criticalSection);
}
