/* CapiHandler.cpp
 *
implementation of the CapiHandler2 class.

Part of the caiviar package. 

Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include "../config.h"
#if defined(LINUX) || defined(CYGWIN)
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <memory.h>
#include "CapiHandler2.h"
#include "Log.h"
#elif defined(WIN32)
#include "CapiHandler2.h"
#include "Log.h"
#include <malloc.h>
#include <time.h>
#include "capi20.h" //?
#ifndef __windows_h_included__
#define __windows_h_included__
#include <windows.h>
#endif
#endif
#include "range.h"

/* if you set this to 1, the capi won't send a alert_req immediately
 * after each connect_ind. This means that other applications are free
 * to take the call as long as we didn't. Unless you're sure what you're
 * doing, set this to 0 */
#define DONT_DO_ALERTING 0

#define MaxNumBChan 1
#define MaxNumB3DataBlocks 7 // 1-7
#define DataBlockSize 512  /* 128-2048 */
#define READ_FORWARD 4  /* must be < 8 | bigger=more stable, smaller=better barge in */
#define CONNECT_READ_FORWARD 2  /* must be < 8, too great value may introduce delays  */

static unsigned char global_job_counter = 1;

#define CONTROLLER(line) (channeltab[line].controller)
#define APPL_ID(line) (channeltab[line].Appl_Id)
#define FLAG_LISTEN(line) (channeltab[line].flag_listen)
#define FLAG_DTMF(line) (channeltab[line].flag_dtmf)
#define FLAG_FAX(line) (channeltab[line].flag_fax)
#define STATE(line) (channeltab[line].state)

//#define DEBUG_DATA_B3

/* these are from ETS 300 102-1
 *
 * you should normally only see are very small subset of these causes.
 * Do not change the simplified mappings unless you're sure what you're doing.
 */
struct reason_t {
    int number;
    char*name;
    int simplified;
}
static reasontab[] = { /*000 0001*/  {1, "Unallocated (unassigned) number", CAPI_REASON_WRONGNUMBER},
                              /*000 0010*/  {2, "No route to specified transit network", CAPI_REASON_TECHNICAL},
                              /*000 0011*/  {3, "No route to destination", CAPI_REASON_TECHNICAL},
                              /*000 0110*/  {6, "channel unacceptable", CAPI_REASON_TECHNICAL},
                              /*000 0111*/  {7, "Call awarded and being delivered in an established channel", CAPI_REASON_TECHNICAL},
                              /*001 0000*/ {16, "Normal call clearing", CAPI_REASON_REJECTED},
                              /*001 0001*/ {17, "User Busy", CAPI_REASON_BUSY},
                              /*001 0010*/ {18, "No User responding", CAPI_REASON_REJECTED},
                              /*001 0011*/ {19, "No answer from user (user alerted)", CAPI_REASON_REJECTED},
                              /*001 0101*/ {21, "Call rejected", CAPI_REASON_REJECTED},
                              /*001 0110*/ {22, "Number changed", CAPI_REASON_WRONGNUMBER},
                              /*001 1010*/ {26, "Nonselected user clearing", CAPI_REASON_TECHNICAL},
                              /*001 1011*/ {27, "Destination out of order", CAPI_REASON_TECHNICAL},
                              /*001 1100*/ {28, "Invalid number format", CAPI_REASON_TECHNICAL},
                              /*001 1101*/ {29, "Faculty rejected", CAPI_REASON_TECHNICAL},
                              /*001 1110*/ {30, "Response to status inqury", CAPI_REASON_TECHNICAL},
                              /*001 1111*/ {31, "Normal, unspecified", CAPI_REASON_TECHNICAL},
                              /*010 0010*/ {34, "No circuit/channel available", CAPI_REASON_BUSY},
                              /*010 0110*/ {38, "Network out of order", CAPI_REASON_TECHNICAL},
                              /*010 1001*/ {41, "Temporary failure", CAPI_REASON_TECHNICAL},
                              /*010 1010*/ {42, "Switching equipment congestion", CAPI_REASON_TECHNICAL},
                              /*010 1011*/ {43, "Access information discarded", CAPI_REASON_TECHNICAL},
                              /*010 1100*/ {44, "Requested circuit/channel not available", CAPI_REASON_TECHNICAL},
                              /*010 1111*/ {47, "Resource unavailable, unspecified", CAPI_REASON_TECHNICAL}
              };
static char* supplementary_service_names[] =
    {"Get supported Services",
     "Listen",
     "Hold",
     "Retrieve",
     "Suspend",
     "Resume",
     "ECT (Explicit Call Transfer)",
     "3PTY Begin",
     "3PTY END",
     "CF Activate",
     "CF Deactivate",
     "CF Interrogate Parameters",
     "CF Interrogate Numbers",
     "CD (Call deflection)",
     "MCID (Malicious Call Identification",
     "CCBS Request (Completion of Calls to busy Sub)",
     "CCBS Deactivate",
     "CCBS Interrogate",
     "CCBS Call"
    };
static int supplementary_service_num = sizeof(supplementary_service_names)/sizeof(char*);

static void log_message(int line, Logger*log, _cmsg*msg)
{
    char* a = capi_cmsg2str(msg);
    char* b = a;
    while(1) {
	char*c = b;
	while(*c && *c!=10)
	    c++;
	if(!*c)
	    c=0;
	else 
	    *c = 0;
	log->logf("<trace> [line %d] | %s", line, b);
	if(!c)
	    break;
	b = c+1;
    }
}

Queue::Queue(int size, int nmemb)
{
    this->queue = (char*)malloc(size * nmemb);
    this->size = size;
    this->nmemb = nmemb;
    readpos = 0;
    writepos = 0;
    number = 0;
    initMutex(&mutex);
}

Queue::~Queue()
{
    free(this->queue);
    destroyMutex(&mutex);
}

int Queue::put(void*data)
{
    int tmp = writepos+1;
    tmp%=nmemb;

    lockMutex(&mutex);
    if(tmp==readpos) {
	unlockMutex(&mutex);
	return 0;
    }
    memcpy(&this->queue[writepos*size], data, size);
    writepos = tmp;
    number++;
    unlockMutex(&mutex);
    return 1;
}

int Queue::get(void*data)
{
    lockMutex(&mutex);
    if(writepos == readpos) {
	unlockMutex(&mutex);
	return 0;
    }
    memcpy(data, &this->queue[readpos*size], size);
    readpos++;
    readpos %= nmemb;
    number--;
    unlockMutex(&mutex);
    return 1;
}

int Queue::flush()
{
    int num;
    lockMutex(&mutex);
    num = this->number;
    this->number = 0;
    this->readpos = 0;
    this->writepos = 0;
    unlockMutex(&mutex);
    return num;
}

int Queue::num()
{
    return number;
}


/* getCapiMessage().  Tries to get a message in the specified time.
 * The timeout variable is a reference. It is used to return 
 * how much time was "left" when we got the message.
 */
_cmsg CapiHandler2::getCapiMessage_internal(int line, int& timeout) {
    unsigned int mexchError;   // MESSAGE_EXCHANGE_ERROR
    _cmsg msg;

    time_t start = time(0);
    int timespend;
    int inittime = timeout;

    do {
        time_t now = time(0);
        timespend = (int)difftime(now , start);
        if (timespend >= timeout && inittime>0)
            break;

        //unsigned int waitRet = CAPI20_WAIT_FOR_SIGNAL(APPL_ID(line));
	lockMutex(&driver_mutex);
        mexchError = CAPI_GET_CMSG(&msg, APPL_ID(line));
	unlockMutex(&driver_mutex);
        if ((MESSAGE_EXCHANGE_ERROR)mexchError != CapiNoError &&
                (MESSAGE_EXCHANGE_ERROR)mexchError != CapiReceiveQueueEmpty)
            log->logf("<error> [line %d] Message exchange error %04x", line, mexchError);

        if(mexchError)
            Sleep(33);
    } while (mexchError != 0);

    timeout -= timespend;
    if (timeout<=0 && inittime>0) {
        timeout = 0;
        return msg;
    }

    return msg;
}

struct ThreadData
{
    CapiHandler2*cp;
    int linenumber;
};

THREAD_FUNCTION lineThread(void*_data)
{
    struct ThreadData*data = (struct ThreadData*)_data;
    CapiHandler2*cp = data->cp;
    int line = data->linenumber;
    delete data;
    cp->workerThread(line);
    return 0;
}

// defines for thread_status
#define THREAD_HALTED 0
#define THREAD_HALT_REQUESTED 1
#define THREAD_RUNNING 2

void CapiHandler2::workerThread(int line)
{
    log->logf("<verbose> [line %d] Worker thread started ", line);
    while(1) {
	_cmsg msg;

	if(channeltab[line].thread_status==1)
	    break;

	int timeout = 1;
	msg = getCapiMessage_internal(line, timeout);
	
	if(timeout<=0) {
	    continue;
	}

	if(msg.Command != CAPI_DATA_B3) {
	    /* move the capi message data fields to userspace.
	       (!) the following will break if it's a DATA_B3 message */
	    int len = (msg.m[1]<<8|msg.m[0]);
	    if(len>=128) {
		log->logf ("<skipping message %d,%d (too big: %d bytes)",
			msg.Command, msg.Subcommand, len);
		continue;
	    }
	    CAPI_MESSAGE tmp = (CAPI_MESSAGE)malloc(128);
	    CAPI_CMSG_2_MESSAGE (&msg, tmp);
	    CAPI_MESSAGE_2_CMSG (&msg, tmp);
	    channeltab[line].controlMessages->put(&msg);
	    continue;
	}

#ifdef DEBUG_DATA_B3
	log->logf("<debug> [line %d] got CAPI message %s: 0x%02x, 0x%02x, (Info:%x) PLCI:0x%x", line,
		   capi_cmd2str(msg.Command, msg.Subcommand),
		   msg.Command, msg.Subcommand, msg.AdditionalInfo, msg.adr.adrPLCI
		 );
#endif

	if (IS_DATA_B3_CONF(&msg)) {
	    lockMutex(&channeltab[line].audio_mutex);
	    int datahandle = msg.DataHandle;
	    if (channeltab[line].outputHandler) {
		/* output handlers transfer a new data block every time
		   a conf comes in. */
		transferData(line);
	    }
	    else if (channeltab[line].playjob.bytes_remaining) {	//still sending
		/*log->logf("<debug> [line %d] playjob %d datahandle %d read_forward %d", line,
			  channeltab[line].playjob.datahandle , datahandle , READ_FORWARD);*/
		if( channeltab[line].playjob.datahandle - datahandle < READ_FORWARD)
		    transferData(line);
	    } else { //all data transferred, handle remaining confirmations.
		if(datahandle == channeltab[line].playjob.datahandle) { // last block got confirmed. we're done.
		    msg.m = (CAPI_MESSAGE)malloc(16);
		    channeltab[line].controlMessages->put(&msg);
		}
	    }
	    unlockMutex(&channeltab[line].audio_mutex);
	} else if (IS_DATA_B3_IND(&msg)) {
	    lockMutex(&channeltab[line].audio_mutex);
	    if(channeltab[line].inputHandler) {
		channeltab[line].inputHandler->handleInput((unsigned char*)DATA_B3_IND_DATA(&msg), 
			    DATA_B3_REQ_DATALENGTH(&msg));
		/* doing record jobs and having an input handler are not
		   mutually exclusive, so we just pass the data on */
	    }

	    if( channeltab[line].recordjob.bytes_remaining>0 ) {
		int size = DATA_B3_REQ_DATALENGTH(&msg);
		if(size > channeltab[line].recordjob.bytes_remaining)
		    size = channeltab[line].recordjob.bytes_remaining;
		log->logf("<notice> [line %d] recording - left:%d", line, channeltab[line].recordjob.bytes_remaining);
		memcpy(channeltab[line].recordjob.data_pos, (void*)DATA_B3_IND_DATA(&msg), size);
		channeltab[line].recordjob.data_size += size;
		channeltab[line].recordjob.data_pos += size;
		channeltab[line].recordjob.bytes_remaining -= size;
		if(channeltab[line].recordjob.bytes_remaining == 0 && size>0) {
		    /* inform the other thread that recording is complete */
		    msg.m = (CAPI_MESSAGE)malloc(16);
		    channeltab[line].controlMessages->put(&msg);
		}
	    }

	    /* If this block was executed any earlier, the data buffer would be freed */
	    _cmsg resp;
	    DATA_B3_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
	    DATA_B3_RESP_NCCI(&resp) = DATA_B3_IND_NCCI(&msg);
	    DATA_B3_RESP_DATAHANDLE(&resp) = DATA_B3_IND_DATAHANDLE(&msg);
	    this->sendCapiMessage(line, &resp);
	    
	    unlockMutex(&channeltab[line].audio_mutex);
	}
    }
    
    log->logf("<verbose> [line %d] Worker thread ended for line", line);
    channeltab[line].thread_status=0;
}

void CapiHandler2::freeMessage(int line, _cmsg* msg)
{
    free(msg->m);
    msg->m = 0;
}

_cmsg CapiHandler2::getMessage(int line, int& ms)
{
    _cmsg msg;
    while(1) {
	if(channeltab[line].delayedMessages->get(&msg)) {
	    log->logf("<verbose> [line %d] reading back message...", line);
	    break;
	}
	if(channeltab[line].controlMessages->get(&msg)) {
	    break;
	}
	Sleep(10);
	ms -= 10;
	if(ms <= 0) {
	    ms = 0;
	    msg.Command = 0;
	    msg.Subcommand = 0;
	    return msg;
	}
    }

    if(msg.AdditionalInfo) {
	log->logf("<verbose> [line %d] CAPI message %s has Info field set: 0x%02x, 0x%02x, (Info:%x) PLCI:0x%x", line,
		capi_cmd2str(msg.Command, msg.Subcommand),
		msg.Command, msg.Subcommand, msg.AdditionalInfo, msg.adr.adrPLCI
		 );
	log_message(line, log, &msg);
    } else {
	log->logf("<debug> [line %d] got CAPI message %s: 0x%02x, 0x%02x, (Info:%x) PLCI:0x%x", line,
		   capi_cmd2str(msg.Command, msg.Subcommand),
		   msg.Command, msg.Subcommand, msg.AdditionalInfo, msg.adr.adrPLCI
		 );
	log_message(line, log, &msg);
    }
    return msg;
}

/* Register, open, and initialize a specific line.
 * The msisdn is the msisdn under which the line
 * should listen for calls _and_ which should appear
 * for the called side when making calls
 */
int CapiHandler2::openLine(int line, char*msisdn) {
    if(channeltab[line].active) {
        log->logf("<warning> [line %d] Can't register line %d. Already active.", line, line);
        return 0;
    }

    CAPI_REGISTER_ERROR ErrorCode;
    lockMutex(&driver_mutex);
    APPL_ID(line) = CAPI20_REGISTER(MaxNumBChan, MaxNumB3DataBlocks, DataBlockSize, &ErrorCode);
    unlockMutex(&driver_mutex);
    if (APPL_ID(line) == 0) {
        log->logf("<fatal> [line %d] registering the application in function RegisterCAPI, ", line);
        log->logf("<fatal> [line %d] code: %04X", line,ErrorCode);
        return 0;
    }
    log->logf("<verbose> [line %d] Registered as BChannel %d of Controller %d, ApplicationID: %d",
	    line, channeltab[line].controller_line, channeltab[line].controller, 
	    APPL_ID(line));

    channeltab[line].active = 1;
    channeltab[line].my_msisdn = strdup(msisdn);
    channeltab[line].inputHandler = 0;
    channeltab[line].outputHandler = 0;

    initMutex(&channeltab[line].audio_mutex);

    FLAG_LISTEN(line) = 0;
    resetLineToIdle(line);
    channeltab[line].flag_hold = 0;
    ThreadData*data = new ThreadData();
    data->cp = this;
    data->linenumber = line;
    
    channeltab[line].controlMessages = new Queue(sizeof(_cmsg), 64);
    channeltab[line].delayedMessages = new Queue(sizeof(_cmsg), 64);
    channeltab[line].tmp = new Queue(sizeof(_cmsg), 64);

    log->logf("<verbose> [line %d] Creating worker thread", line);
    channeltab[line].thread_status=2;
    createThread(&channeltab[line].thread, lineThread, data);
    return 1;
}

/* Reset, close and Deregister a specific line.
 */
void CapiHandler2::close(int line) {
    if(!(line>=0 && line<maxchannels && channeltab[line].active))
	return;
    if(channeltab[line].flag_3pty)
	this->endconference(line);

    if(STATE(line) == state_connected || 
       STATE(line) == state_holdreq ||
       STATE(line) == state_3ptyreq)
	this->hangup(line);

    /* in case of a call on hold or 3 party conference, there
       are two lines to be freed, not just one */

    if(channeltab[line].flag_hold) /*
	There's a second line waiting
    */ {
	channeltab[line].plci_ncci = channeltab[line].flag_hold;
	channeltab[line].flag_b3 = 0;
	this->hangup(line);
	channeltab[line].flag_hold = 0;
    }

    if(channeltab[line].thread_status) {
	channeltab[line].thread_status = 1;
	log->logf("<verbose> [line %d] killing worker Thread for line %d", line, line);
	while(channeltab[line].thread_status)
	    Sleep(33);
    }

    log->logf("<notice> [line %d] Releasing line %d (Appl %d)", line, line, APPL_ID(line));
    
    lockMutex(&driver_mutex);
    MESSAGE_EXCHANGE_ERROR ErrorCode;
    ErrorCode = CAPI20_RELEASE(APPL_ID(line));
    unlockMutex(&driver_mutex);
    if (ErrorCode != 0) {
	log->logf("<error> [line %d] deregistering the application in function close(), ", line);
	log->logf("<error> [line %d] code: 0x%04X", line,ErrorCode);
    }

    free(channeltab[line].my_msisdn);
    channeltab[line].my_msisdn = 0;

    destroyMutex(&channeltab[line].audio_mutex);

    channeltab[line].active = 0;

    delete channeltab[line].controlMessages;
    channeltab[line].controlMessages = 0;
    delete channeltab[line].delayedMessages;
    channeltab[line].delayedMessages = 0;
    delete channeltab[line].tmp;
    channeltab[line].tmp = 0;
}

/* Test if a line number is valid. Used in the SimpleCapi implementation
 * to catch some obvious usage-errors.
 */
int CapiHandler2::checkLine(int line) {
    if((line<0) || (line>=maxchannels) || (channeltab[line].active==0)) {
        log->logf("<fatal> [line %d] line %d is invalid. (did you forget to reopen a closed line?)", line, line);
        return 0;
    }
    return 1;
}

/* Send a message to the capi interface.  */
void CapiHandler2::sendCapiMessage(int line, _cmsg* msg) {
#ifndef DEBUG_DATA_B3
    if(msg->Command != CAPI_DATA_B3)
#endif
    {
	log->logf("<debug> [line %d] send CAPI message %s: 0x%02x, 0x%02x, PLCI:0x%x", line,
		capi_cmd2str(msg->Command, msg->Subcommand), msg->Command, msg->Subcommand, msg->adr.adrPLCI);
	log_message(line, log, msg);
    }
    lockMutex(&driver_mutex);
    unsigned int retval = CAPI_PUT_CMSG(msg);
    unlockMutex(&driver_mutex);
    if (retval != 0) {
	log->logf("<error> [line %d] Couldn't send message %s: 0x%02x, 0x%02x, PLCI:0x%x", line,
		  capi_cmd2str(msg->Command, msg->Subcommand),
		  msg->Command, msg->Subcommand, msg->adr.adrPLCI);
    }
}

/* Constructor. Intitialize mutex(s) and the channeltab.
 */
CapiHandler2::CapiHandler2(Logger*log) {
    this->log = log;
    this->allcalls_hack = 0;
    log->log ("<verbose> CapiHandler2::CapiHandler2()");
    initMutex(&driver_mutex);
    initMutex(&capi_mutex);
    init();
}

/* Destructor. close all lines that are open.
 */
CapiHandler2::~CapiHandler2() {
    log->log ("<verbose> CapiHandler2::~CapiHandler2()");
    int t;
    for(t=0;t<maxchannels;t++) {
        close(t);
    }
    destroyMutex(&capi_mutex);
    destroyMutex(&driver_mutex);
}

/* init(). Called from the constructor. Initializes the
 * capi interface and set's up the channeltab.
 */
int CapiHandler2::init() {

    unsigned int numController;

    if (!CAPI20_ISINSTALLED()) {
        log->logf("<fatal> InitCapi: ISDN-driver not installed");
        return 0;
    }

    // get number of controllers
    unsigned short Buffer[64 / sizeof (unsigned short)];
    CAPI20_GET_PROFILE (0, (unsigned char *)Buffer);
    numController = Buffer[0];

    if (numController == 0) {
        log->logf("<fatal> InitCapi: No ISDN-controller installed");
        return 0;
    }

    if (numController == 0)
        return 0;
    log->logf("<notice> %d Controller%s found", numController, numController>=2?"s":"");

    unsigned int lines = 0;
    int globlines = 0;
    memset(channeltab,0,sizeof(channeltab));
    for (unsigned int i=1; i<=numController; i++) {
        unsigned char buf[12];

        CAPI20_GET_SERIAL_NUMBER(buf);

	unsigned short Buffer[64 / sizeof (unsigned short)];

	/* get number of b channels */
	CAPI20_GET_PROFILE((unsigned)i, (unsigned char *)Buffer);
	// Buffer actually contains a lot of useful information
	// we might be interested in.
	lines = (unsigned)Buffer[1];

        log->logf("<notice> Controller #%d (%s) supports %d lines", i, buf, lines);

        unsigned int j;
        for(j=0;j<lines;j++) {
            channeltab[globlines].active = 0;
            channeltab[globlines].controller = i;
            channeltab[globlines].controller_line = j;
            channeltab[globlines].other_msisdn[0] = 0; // set msisdn to ""
            channeltab[globlines].playjob.bytes_remaining = 0;
            channeltab[globlines].recordjob.bytes_remaining = 0;
	    channeltab[globlines].controlMessages = 0;
	    channeltab[globlines].delayedMessages = 0;
	    channeltab[globlines].tmp = 0;

            globlines ++;
        }
    }
    maxchannels = globlines;

    return maxchannels;
}

/* Reset a line to idle state. Notice! This just clears
 * all variables. It doesn't restore any physical states.
 * This routine does _not_ reset the field flag_hold!
 */
void CapiHandler2::resetLineToIdle(int line) {
    STATE(line) = state_idle;
    lockMutex(&channeltab[line].audio_mutex);
    if(channeltab[line].playjob.bytes_remaining) {
        free(channeltab[line].playjob.data);
        channeltab[line].playjob.bytes_remaining = 0;
    }
    
    if(channeltab[line].recordjob.bytes_remaining) {
	struct job*j = &channeltab[line].recordjob;
	channeltab[line].recordjob.bytes_remaining = 0;
	if(j->callback) {
	    int size = j->data_size;
	    unlockMutex(&channeltab[line].audio_mutex);
	    j->callback(j->callback_data, size);
	    j->callback = 0;
	}
    } else {
	unlockMutex(&channeltab[line].audio_mutex);
    }

    if(channeltab[line].controlMessages) {
	log->logf("<verbose> [line %d] Resetting line. %d unprocessed messages in messagequeue.", line, 
		channeltab[line].delayedMessages->flush());
    }

    FLAG_DTMF(line) = 0;
    FLAG_FAX(line) = 0;
    channeltab[line].flag_3pty = 0;
    channeltab[line].flag_b3 = 0;
   // channeltab[line].flag_hold = 0;
}

/* Open a free line, and assign it to the given msisdn.
 * If there's no free line, returns -1.
 */
int CapiHandler2::open (char*msisdn, char*controller) {
    int t;
    log->logf("<verbose> CapiHandler2::open(%s, %s)", msisdn, controller);
    for(t=0;t<maxchannels;t++) {
        if(channeltab[t].active == 0
	   && is_in_range(channeltab[t].controller, controller)) {
            if(!openLine(t, msisdn))
                return -1;

            return t;
        }
    }
    log->logf("<warning> No free line in range %s!", controller);
    return  -1;
}

/* acceptCalls. Notifies the capi interface that the caller wants to get notified
 * of incoming calls.
 */
void CapiHandler2::acceptCalls(int line) {
    FLAG_LISTEN(line) = 1;
    // listen to incoming calls
    _cmsg msg;
    LISTEN_REQ_HEADER(&msg, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    LISTEN_REQ_CONTROLLER(&msg) = CONTROLLER(line);
    LISTEN_REQ_INFOMASK(&msg) = 128; //setting this to 0 would probably stop those extra INFO_IND messages
    LISTEN_REQ_CIPMASK(&msg) = 0x1FFF03FF;   // ALL_SERVICES
    LISTEN_REQ_CIPMASK2(&msg) = 0;
    LISTEN_REQ_CALLINGPARTYNUMBER(&msg) = NULL;
    LISTEN_REQ_CALLINGPARTYSUBADDRESS(&msg) = NULL;
    log->logf("<verbose> [line %d] Sending LISTEN_REQ...", line);
    sendCapiMessage(line, &msg);
    
    log->logf("<verbose> [line %d] Waiting for LISTEN_CONF...", line);

    // wait for LISTEN_CONF
    msg = waitForEvent(line, CAPI_LISTEN, CAPI_CONF, 1);

    if (LISTEN_CONF_INFO(&msg) != 0) {
        log->logf ("<warning> [line %d] error at registering listener: 0x%02x", line, LISTEN_CONF_INFO(&msg));
    } else {
	log->logf("<verbose> [line %d] ...done", line);
	log->logf("<verbose> [line %d] CAPI listener registered correctly", line);
    }
    freeMessage(line, &msg);
}

/* wait until a specified message comes in. E.g. event=0x00, subevent=0x82 can be used
 * to wait for the next IND message
 */
_cmsg CapiHandler2::waitForEvent(int line, int event, int subevent, int store) //0 == joker!
{
    _cmsg msg,tmp;
    log->logf("<debug> [line %d] Blocking wait for 0x%02x 0x%02x", line, event, subevent);
    while (true) {
	int timeout = 3600*1000;
        msg = getMessage(line, timeout);
        if(timeout > 0) {
            if ((msg.Command == event || event == 0 ) &&
                    (msg.Subcommand == subevent || subevent == 0)) {
                log->logf("<debug> [line %d] Blocking wait for 0x%02x 0x%02x ended.", line,event, subevent);
                break;
            } else {
                log->logf("<debug> [line %d] [wait for 0x%02x 0x%02x] unexpected reply %s: 0x%02x 0x%02x %s", line,
                          event, subevent,
                          capi_cmd2str(msg.Command, msg.Subcommand),
                          msg.Command, msg.Subcommand,
                          store?"(storing)":""
                         );
                if(store)
                    channeltab[line].tmp->put(&msg);
		else
		    freeMessage(line, &msg);
            }
        }
    }
    while(channeltab[line].tmp->get(&tmp))
	channeltab[line].delayedMessages->put(&tmp);
    return msg;
}

/* use a (already openend) line to call a specified msisdn.
 * mode < 0: normal call
 * mode == 0: fax sff
 * mode == 5: fax ascii
 */
int CapiHandler2::call (int line, char* msisdn, int mode) {
    unsigned int len1 = strlen((char*)msisdn);
    unsigned char* pCalledNumberStruct = (unsigned char *)malloc(len1 + 2 + 1); // gets freed
    pCalledNumberStruct[0] = (unsigned char)(len1 + 1);
    pCalledNumberStruct[1] = 0x80;
    strcpy ((char*)&(pCalledNumberStruct[2]), (char*)msisdn);

    char*mynumber = channeltab[line].my_msisdn;
    unsigned int len2 = (mynumber == NULL)? 1 : strlen((char*)mynumber);
    unsigned char* pCallingNumberStruct = (unsigned char *)malloc(len2 + 2 + 1); // gets freed
    if (mynumber != NULL) {
        pCallingNumberStruct[0] = (unsigned char)(len2 + 1);
        pCallingNumberStruct[1] = 0x80;
        strcpy ((char*)&(pCallingNumberStruct[2]), (char*)mynumber);
    } else {
        // no msn supplied, fill in basic values
        pCallingNumberStruct[0] = 0x02;
        pCallingNumberStruct[1] = 0x00;
        pCallingNumberStruct[2] = 0x80;
    }

    int b1proto = 1; // 64 kbit with byte framing
    int b2proto = 1; // Transparent
    int b3proto = 0; // Transparent

    B3_PROTO_FAXG3  B3conf;
    if(mode >= 0) {
        b1proto = 4; // T.30 modem for Group 3 fax
        b2proto = 4; // T.30 modem for Group 3 fax
        b3proto = 4; // T.30 for Group 3 fax
        B3conf.resolution = 0;
        B3conf.format     = (unsigned short)mode;
        char    stationID[30] = "45";
        char    headLine [30] = "Simple Fax Page";
        len1 = strlen( stationID );
        B3conf.Infos[0] = (unsigned char)len1;
        strcpy( (char *)&B3conf.Infos[1], stationID );
        len2 = strlen( headLine );
        B3conf.Infos[len1 + 1] = (unsigned char)len2;
        strcpy( (char *)&B3conf.Infos[len1 + 2], headLine );
        B3conf.len = (unsigned char)(2 * sizeof( unsigned short ) + len1 + len2 + 2);
    }

    _cmsg msg;
    CONNECT_REQ_HEADER(&msg, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    CONNECT_REQ_B1PROTOCOL(&msg) = b1proto;
    CONNECT_REQ_B2PROTOCOL(&msg) = b2proto;
    CONNECT_REQ_B3PROTOCOL(&msg) = b3proto;
    CONNECT_REQ_CONTROLLER(&msg) = CONTROLLER(line);
    CONNECT_REQ_CIPVALUE(&msg) = 1;
    CONNECT_REQ_CALLEDPARTYNUMBER(&msg) = pCalledNumberStruct;
    CONNECT_REQ_CALLINGPARTYNUMBER(&msg) = pCallingNumberStruct;
    
    if(mode >= 0) {
        CONNECT_REQ_B3CONFIGURATION(&msg) = (_cstruct)&B3conf;
    }

    sendCapiMessage(line, &msg);

    free(pCallingNumberStruct);
    free(pCalledNumberStruct);

    // wait for CONNECT_CONF
    _cmsg msg2 = waitForEvent (line, CAPI_CONNECT, CAPI_CONF, 1);
    channeltab[line].plci_ncci = msg2.adr.adrPLCI;
    freeMessage(line, &msg2);

    STATE(line) = state_calling;
    FLAG_FAX(line) = (mode>=0)? 1 : 0;

    strncpy(channeltab[line].other_msisdn, msisdn, sizeof(channeltab[line].other_msisdn));
    channeltab[line].callingjob = 0x0300 | line;
    return channeltab[line].callingjob;
}

/* find out which sound file is currently playing, send the next datablock
 * to capi, and update the playjob structure 
 */
void CapiHandler2::transferData(int line) {
    _cmsg  msg;
    
    if(channeltab[line].outputHandler) {
	int size = DataBlockSize;
	unsigned char data[DataBlockSize];
	size = channeltab[line].outputHandler->handleOutput(data,size);
	if(!size) {
	    size = DataBlockSize/2;
	    log->logf("<verbose> [line %d] transferData: handleOutput returned zero size- filling up to %d", line, size);
	    int t;
	    for(t=0;t<size;t++) {
		if(t&1) data[t] = 0x7f^0x55;
		else data[t] = 0xff^0x55;
	    }
	}

        DATA_B3_REQ_HEADER(&msg, APPL_ID(line), 1234, CONTROLLER(line));
	DATA_B3_REQ_NCCI(&msg) = channeltab[line].plci_ncci;
#if defined(WIN32) || defined(CYGWIN)
        DATA_B3_REQ_DATA(&msg) = (unsigned long)data;
#elif defined(LINUX)
        DATA_B3_REQ_DATA(&msg) = (unsigned char*)(unsigned long)data;
#endif
        DATA_B3_REQ_DATALENGTH(&msg) = size;
        DATA_B3_REQ_DATAHANDLE(&msg) = channeltab[line].playjob.datahandle;
        DATA_B3_REQ_FLAGS(&msg) |= 2; //more data //new
        channeltab[line].playjob.datahandle++;
        sendCapiMessage(line, &msg);
	return;
    }

    if (channeltab[line].playjob.bytes_remaining > 0) {
        int size = DataBlockSize;
        if(size > channeltab[line].playjob.bytes_remaining)
            size = channeltab[line].playjob.bytes_remaining;

        //log->logf("<verbose> [line %d] Sending datahandle %d", line, channeltab[line].playjob.datahandle);

        DATA_B3_REQ_HEADER(&msg, APPL_ID(line), 1234, CONTROLLER(line));
        msg.adr.adrNCCI = channeltab[line].plci_ncci;
#if defined(WIN32) || defined(CYGWIN)
        DATA_B3_REQ_DATA(&msg) = (unsigned long)channeltab[line].playjob.data_pos;
#elif defined(LINUX)
        DATA_B3_REQ_DATA(&msg) = (unsigned char*)(unsigned long)channeltab[line].playjob.data_pos;
#endif
        DATA_B3_REQ_DATALENGTH(&msg) = size;
        DATA_B3_REQ_DATAHANDLE(&msg) = channeltab[line].playjob.datahandle;
        DATA_B3_REQ_FLAGS(&msg) |= 2; //more data //new

        sendCapiMessage(line, &msg);
	{
            channeltab[line].playjob.data_pos += size;
            channeltab[line].playjob.bytes_remaining -= size;
            if(channeltab[line].playjob.bytes_remaining == 0) {
                free(channeltab[line].playjob.data);
            } else
                channeltab[line].playjob.datahandle++;
        }
	return;
    }
}

/* called from SimpleCapi- starts playing a sound file
 * (must be a .law)
 */
int CapiHandler2::play (int line, void*mem, int len) {
    lockMutex(&channeltab[line].audio_mutex);
    if(channeltab[line].outputHandler) {
        log->logf("<notice> [line %d] play() called: killing Output Handler", line);
	channeltab[line].outputHandler = 0;
    }
    if(channeltab[line].playjob.bytes_remaining) {
        log->logf("<error> [line %d] play: Device busy. (Stop the other job first)", line);
	unlockMutex(&channeltab[line].audio_mutex);
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("<error> [line %d] play: Not connected.", line);
	unlockMutex(&channeltab[line].audio_mutex);
        return 0;
    }
    int jobnr = global_job_counter++;
    channeltab[line].playjob.data = (char*)malloc (len);
    memcpy (channeltab[line].playjob.data, mem , len);
    channeltab[line].playjob.data_pos = channeltab[line].playjob.data;
    channeltab[line].playjob.data_size = len;
    channeltab[line].playjob.bytes_remaining = len;
    channeltab[line].playjob.datahandle = 256*jobnr;

    jobnr |= 0x0100;

    channeltab[line].playjob.jobid = jobnr;

    int t;
    for(t=0;t<READ_FORWARD;t++)
        transferData(line);
    unlockMutex(&channeltab[line].audio_mutex);
    
    return jobnr;
}

/* Start recording B3 data.
 */
int CapiHandler2::record (int line, void*buf, int bufsize, void(*callback)(void*,int),void*callback_data) {
    if(channeltab[line].recordjob.bytes_remaining) {
        log->logf("<error> [line %d] record: Device busy. (Stop the other job first)", line);
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("<error> [line %d] record: Not connected.", line);
        return 0;
    }
    int jobnr = (global_job_counter++)&255;
    channeltab[line].recordjob.data = (char*)buf;
    channeltab[line].recordjob.data_pos = (char*)buf;
    channeltab[line].recordjob.data_size = 0;
    channeltab[line].recordjob.bytes_remaining = bufsize;
    channeltab[line].recordjob.datahandle = 256*jobnr;
    jobnr |= 0x0200;
    channeltab[line].recordjob.jobid = jobnr;
    channeltab[line].recordjob.callback = callback;
    channeltab[line].recordjob.callback_data = callback_data;
    return jobnr;
}

/* will report the number of bytes received over isdn in the 
 * last record attempt
 */
int CapiHandler2::getRecordSize(int line) {
    return channeltab[line].recordjob.data_size;
}

/* will attempt to stop play jobs, record jobs and call jobs. */
void CapiHandler2::stop (int line, int jobid) {
    if((jobid&0xff00)==0x0100) //playjob
    {
	lockMutex(&channeltab[line].audio_mutex);
        if(channeltab[line].playjob.bytes_remaining) {
            free(channeltab[line].playjob.data);
            channeltab[line].playjob.bytes_remaining = 0;
            channeltab[line].playjob.jobid = 0;
        }
	unlockMutex(&channeltab[line].audio_mutex);
    } else if((jobid&0xff00)==0x0200) //recordjob
    {
	lockMutex(&channeltab[line].audio_mutex);
        if(channeltab[line].recordjob.bytes_remaining) {
            channeltab[line].recordjob.bytes_remaining = 0;
            channeltab[line].recordjob.jobid = 0;
        }
	unlockMutex(&channeltab[line].audio_mutex);
    } else if((jobid&0xff00)==0x0300) //call job
    {
        _cmsg msg;
	if(channeltab[line].state != state_calling) {
	    log->logf("<error> [line %d] stop() called for call while not in calling state!", line);
	    return;
	}
        DISCONNECT_REQ_HEADER(&msg, APPL_ID(line), 1235, CONTROLLER(line));
        DISCONNECT_REQ_PLCI(&msg) = channeltab[line].plci_ncci;
        sendCapiMessage(line, &msg);
	msg = waitForEvent(line, CAPI_DISCONNECT, CAPI_CONF, 1);
	freeMessage(line, &msg);
    }
}

/* capi gives us a reason when a call didn't make
 * it through. this routine resolves the
 * reason code, print out the verbose meaning,
 * and returns a simplified version
 */
int CapiHandler2::translatereason(int line, int reason) {
    int simplified = 0;
    if(reason >= 0x3301 && reason <= 0x3303) {
        log->logf("<verbose> [line %d] Protocol error %04x", line, reason);
        simplified = CAPI_REASON_TECHNICAL;
    } else if(reason == 0x3304) {
        log->logf("<verbose> [line %d] Another application got that call", line);
        simplified = CAPI_REASON_TECHNICAL;
    } else if((reason & 0xff00) == 0x3400) {
        int t;
        int s=-1;
        for(t=0;t<(sizeof(reasontab)/sizeof(reason_t));t++) {
            if(reasontab[t].number == (reason & 0x7f))
                s=t;
        }
        if(s>=0) {
            simplified = reasontab[s].simplified;
            log->logf("<verbose> [line %d] reason:%s", line,reasontab[s].name);
        } else {
            simplified = CAPI_REASON_TECHNICAL;
            log->logf("<verbose> [line %d] unknown 34xx reason: %04x", line,reason);
        }
    } else {
        simplified = CAPI_REASON_TECHNICAL;
	if(reason==0)
	    log->logf("<verbose> [line %d] clearing in accordance with protocol: %04x", line,reason);
	else
	    log->logf("<verbose> [line %d] unknown reason: %04x", line,reason);
    }
    return simplified;
}

/* supplementary services have their own reason codings.
 * This routine just tries to translate them into english
 * and log them. */
void CapiHandler2::showSupplementaryReason(int line, short int reason) {
    log->logf("<verbose> [line %d] supplementary service reason: %04x", line, reason);
    if(reason == 0x3303)
        log->logf("<verbose> [line %d] (network did not answer in time)", line);
}


/* handles the _cmsg, in that it sends RESP messages, performs state
 * transitions, and, if an ISDNEvent occured, copies it to the given
 * pointer  and returns 1 */
int CapiHandler2::handleMessage(int line, _cmsg mymsg, ISDNEvent*event) {
    //log->logf("<verbose> [line %d] handling message 0x%02x 0x%02x\n", line, mymsg.Command, mymsg.Subcommand);

    if (IS_CONNECT_IND(&mymsg)) {
        if( STATE(line) == state_connected) {
            // we are busy. we ignore this call, maybe some other
            // application wants to take it.

            CONNECT_RESP_HEADER(&mymsg, APPL_ID(line), mymsg.Messagenumber, CONTROLLER(line));
            CONNECT_RESP_PLCI(&mymsg) = CONNECT_IND_PLCI(&mymsg);
            CONNECT_RESP_REJECT(&mymsg) = 1;  // 1: ignore call
            CONNECT_RESP_B1PROTOCOL(&mymsg) = 1;
            CONNECT_RESP_B2PROTOCOL(&mymsg) = 1;
            CONNECT_RESP_B3PROTOCOL(&mymsg) = 0;
            this->sendCapiMessage(line, &mymsg);
	    log->logf("<verbose> [line %d] sending CONNECT_RESP-ignore", line);
            return 0;
        }
        if (!FLAG_LISTEN(line)) // not awaiting calls
        {
            log->logf("<notice> [line %d] rejecting call", line);
            // send CONNECT_RESP
            CONNECT_RESP_HEADER(&mymsg, APPL_ID(line), mymsg.Messagenumber, CONTROLLER(line));
            CONNECT_RESP_PLCI(&mymsg) = CONNECT_IND_PLCI(&mymsg);
            CONNECT_RESP_REJECT(&mymsg) = 2;  // 2: reject call, normal call clearing
            CONNECT_RESP_B1PROTOCOL(&mymsg) = 1;
            CONNECT_RESP_B2PROTOCOL(&mymsg) = 1;
            CONNECT_RESP_B3PROTOCOL(&mymsg) = 0;
            this->sendCapiMessage(line, &mymsg);
            return 0;
        } else {
            if(DONT_DO_ALERTING) {
                channeltab[line].plci_ncci = CONNECT_IND_PLCI(&mymsg);
                channeltab[line].incoming = (_cmsg*)malloc(sizeof(mymsg));
                memcpy(channeltab[line].incoming, (void*)&mymsg, sizeof(mymsg));
                STATE(line) = state_alerted;
                event->line = line;
                event->type = CAPI_EVENT_INCOMING;
                event->msisdn = "";
                event->called_msisdn = "";
                event->reason = 0;
                event->internal = channeltab[line].incoming;
                /*return*/
                return 1;
            }

            unsigned char *number = CONNECT_IND_CALLEDPARTYNUMBER(&mymsg);
            char * called =  (char*)malloc(number[0]+1);
            if(number[0]>=2) {
                strncpy(called, (const char*)(&number[2]), number[0]-1);
		called[number[0]-1] = 0;
		log->logf("<verbose> [line %d] incoming call to %s", line, called);
	    } else {
		called[0] = 0;
		log->logf("<verbose> [line %d] incoming call to ?", line);
	    }

	    if(!allcalls_hack && channeltab[line].my_msisdn)
	    if(number[0]>1 && strcmp(called, channeltab[line].my_msisdn)) {
		CONNECT_RESP_HEADER(&mymsg, APPL_ID(line), mymsg.Messagenumber, CONTROLLER(line));
		CONNECT_RESP_PLCI(&mymsg) = CONNECT_IND_PLCI(&mymsg);
		CONNECT_RESP_REJECT(&mymsg) = 1;  // 1: ignore call
		CONNECT_RESP_B1PROTOCOL(&mymsg) = 1;
		CONNECT_RESP_B2PROTOCOL(&mymsg) = 1;
		CONNECT_RESP_B3PROTOCOL(&mymsg) = 0;
		this->sendCapiMessage(line, &mymsg);
		log->logf("<verbose> [line %d] sending CONNECT_RESP(ignore) (wrong number: (%d/%02x/\"%s\"{%d} != \"%s\"{%d}))", line, 
			number[0],
			number[1], 
			called, strlen(called),
			channeltab[line].my_msisdn, strlen(channeltab[line].my_msisdn));
		free(called);
		return 0;
	    }

            // send ALERT_REQ
            _cmsg resp;
            ALERT_REQ_HEADER(&resp,mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
            ALERT_REQ_PLCI(&resp) = CONNECT_IND_PLCI(&mymsg);
            this->sendCapiMessage(line, &resp);
            STATE(line) = state_alertreq;

            char * calling = (char*)malloc(CONNECT_IND_CALLINGPARTYNUMBER(&mymsg)[0]+1);
            if(CONNECT_IND_CALLINGPARTYNUMBER(&mymsg)[0] <= 2)
                calling[0] = 0;
            else
                sprintf (calling, "%s", &CONNECT_IND_CALLINGPARTYNUMBER(&mymsg)[3]);
            strncpy(channeltab[line].other_msisdn, (const char*)calling, sizeof(channeltab[line].other_msisdn));
	    strncpy(channeltab[line].called_msisdn, (const char*)called, sizeof(channeltab[line].called_msisdn));

            log->logf("<notice> [line %d] %s calling %s- connecting...", line, calling, called);
            channeltab[line].incoming = (_cmsg*)malloc(sizeof(mymsg));
            memcpy(channeltab[line].incoming, (void*)&mymsg, sizeof(mymsg));

            unsigned char * calledsub = CONNECT_IND_CALLEDPARTYSUBADDRESS(&mymsg);
            unsigned char *callingsub = CONNECT_IND_CALLINGPARTYSUBADDRESS(&mymsg);

            free(calling);
            free(called);
            return 0;
        }
    } else if (IS_ALERT_CONF(&mymsg)) {
        channeltab[line].plci_ncci = ALERT_CONF_PLCI(&mymsg);
        if (STATE(line) == state_alertreq) {
            STATE(line) = state_alerted;
            log->logf("<verbose> [line %d] got ALERT_CONF", line);

	    if(channeltab[line].outputHandler) {
		/* if the application has set an outputHandler, we now start
		   sending data. (I.e., send the first block- the next blocks
		   will be sent whenever a confirmation message comes in)
		 */
		int t;
		lockMutex(&channeltab[line].audio_mutex);
		for(t=0;t<CONNECT_READ_FORWARD;t++)
		    transferData(line);
		unlockMutex(&channeltab[line].audio_mutex);
	    }

            // we got the call into alert state. now
            // it's time to report the fact that someone is calling
            // back to our caller.
            event->line = line;
            event->type = CAPI_EVENT_INCOMING;
            event->msisdn = channeltab[line].other_msisdn;
            event->called_msisdn = channeltab[line].called_msisdn;
            event->reason = 0;
            event->internal = channeltab[line].incoming;
            /*return*/
            return 1;

        } else {
            log->logf("<error> [line %d] got unexpected ALERT_CONF", line);
            return 0;
        }
    } else if (IS_CONNECT_ACTIVE_IND(&mymsg)) {
        log->logf("<verbose> [line %d] got connect active ind", line);
        _cmsg resp;
        // send CONNECT_ACTIVE_RESP
        CONNECT_ACTIVE_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        CONNECT_ACTIVE_RESP_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&mymsg);
        sendCapiMessage(line, &resp);
        // send CONNECT_B3_REQ
        CONNECT_B3_REQ_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_REQ_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&mymsg);
        CONNECT_B3_REQ_NCPI(&resp) = NULL;
        
        sendCapiMessage(line, &resp);
        return 0;
    } else if (IS_CONNECT_B3_ACTIVE_IND(&mymsg)) {
        log->logf("<verbose> [line %d] got connect b3 active ind", line);
        _cmsg resp;
        // send CONNECT_B3_ACTIVE_RESP
        log->logf("<debug> [line %d] sending b3 active response", line);
        CONNECT_B3_ACTIVE_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_ACTIVE_RESP_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&mymsg);
        sendCapiMessage(line, &resp);

        channeltab[line].plci_ncci = CONNECT_B3_ACTIVE_IND_NCCI(&mymsg);
        log->logf("<verbose> [line %d] Setting PLCI/NCCI to %x", line, channeltab[line].plci_ncci);

        // try to enable dialtones
        if(!channeltab[line].flag_fax) {
            FACILITY_REQ_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
            FACILITY_REQ_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&mymsg);
            FACILITY_REQ_FACILITYSELECTOR(&resp) = (unsigned short)1; // 1: DTMF
            /*len  listen  msdigit mspause tosend  characters */
            unsigned char cBuffer[] = {8,  1,0,    40,0,   40,0,   0,      0,         0};
            FACILITY_REQ_FACILITYREQUESTPARAMETER(&resp) = cBuffer;
            sendCapiMessage(line, &resp);
        }

        STATE(line) = state_connected;
	channeltab[line].flag_b3 = 1;

        event->line = line;
        event->type = CAPI_EVENT_CONNECTED;
        event->msisdn = channeltab[line].other_msisdn;
        event->called_msisdn = channeltab[line].called_msisdn;
        event->reason = 0;
        event->jobid = channeltab[line].callingjob;
        event->tone = -1;
        /*return*/
        return 1;
    } else if (IS_CONNECT_B3_CONF(&mymsg)) {
        log->logf("<verbose> [line %d] got connect b3 conf", line);
        return 0;
    } else if (IS_FACILITY_CONF(&mymsg)) {
        unsigned char*par = FACILITY_CONF_FACILITYCONFIRMATIONPARAMETER(&mymsg);

        if (FACILITY_CONF_FACILITYSELECTOR(&mymsg) == 3) {
            log->logf("<verbose> [line %d] got facility conf. (selector = 3)", line);
            int len0=par[0];
            if(len0>3 && par[1]==0 && par[2]==0) { //get supplementary services
                int len=par[3];
                if(len>2 && par[4]==0 && par[5]==0) {
                    log->logf("<verbose> [line %d] supported services: %04x", line,*(unsigned long int*)(par+6));
                }
            }
        } else if (FACILITY_CONF_FACILITYSELECTOR(&mymsg) == 1) {
            FLAG_DTMF(line) = 1;
        } else {
            log->logf("<error> [line %d] unknown facility_conf: %d (dialtones not supported?)", line,
                      FACILITY_CONF_FACILITYSELECTOR(&mymsg));
        }
    } else if (IS_FACILITY_IND(&mymsg)) {
        int selector = FACILITY_IND_FACILITYSELECTOR(&mymsg);
        if (selector == 1) //DTMF
        {
            _cstruct indParams = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mymsg);
            // FIXME: if there's more than one, put them in a loop.
            int len = indParams[0];
	    char tone = '\0';
	    if(len>0)
		tone = indParams[1];
            if(len>1)
                log->logf("<error> [line %d] missed %d dtmf(s)!", line, indParams[0]-1);
            for(int i = 1; i <= len; i++) {
                log->logf ( "<notice> [line %d] detected DTMF %d (of %d) is %c %s", line, i, len, indParams[i],
			i>1?"(ignoring)":"");
            }

            _cmsg resp;
            FACILITY_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
            FACILITY_RESP_PLCI(&resp) = FACILITY_IND_PLCI(&mymsg);
            FACILITY_RESP_NCCI(&resp) = FACILITY_IND_NCCI(&mymsg);
            FACILITY_RESP_FACILITYSELECTOR(&resp) = FACILITY_IND_FACILITYSELECTOR(&mymsg);
            FACILITY_RESP_FACILITYRESPONSEPARAMETERS(&resp) = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mymsg);
            sendCapiMessage(line, &resp);

            if(len>0) {
                event->line = line;
                event->type = CAPI_EVENT_DIALTONE;
                event->msisdn = channeltab[line].other_msisdn;
		event->called_msisdn = channeltab[line].called_msisdn;
                event->reason = 0;
                event->jobid = -1;
                event->tone = tone;
                /*return*/
                return 1;
            }
        } else {
            unsigned char*fi = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mymsg);
            log->logf("<verbose> [line %d] Got facility ind (selector = %d)", line, selector);
            int len=fi[0];

            _cmsg resp;
            FACILITY_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
            FACILITY_RESP_PLCI(&resp) = FACILITY_IND_PLCI(&mymsg);
            FACILITY_RESP_NCCI(&resp) = FACILITY_IND_NCCI(&mymsg);
            FACILITY_RESP_FACILITYSELECTOR(&resp) = FACILITY_IND_FACILITYSELECTOR(&mymsg);
            FACILITY_RESP_FACILITYRESPONSEPARAMETERS(&resp) = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mymsg);
            sendCapiMessage(line, &resp);
           
/*	    int t;
	    for(t=0;t<len;t++)
            	log->logf("%02x ",fi[t+1]);*/

            if(len>=3) {
                int function = *(short int*)&fi[1];
                int len2 = fi[3];
                int reason = -1;
		/*
                for(t=0;t<len2;t++)
                  log->logf("%02x",fi[t+4]);*/
                if(len2>=2) {
                    reason = *(short int*)&fi[4];
                    showSupplementaryReason(line, reason);
                }

                if(function < supplementary_service_num)
                    log->logf("<verbose> [line %d] Indicator is for Function 0x%02x (%s)", line, function,
                              supplementary_service_names[function]);
                else
                    log->logf("<error> [line %d] Indicator is for Function 0x%02x (unknown service)", line, function);

                if(!reason)
                    switch(function) {
                    case 0x0d: // CD
                        break;
                    case 0x09: // CF
                        if(len2>=6) {
                            int reason = *(short int*)&fi[4];
                            int handle = *(long int*)&fi[6];
                            log->logf("<verbose> [line %d] supplementary service handle: %08x", line, handle);
                        }
                        break;
                    case 0x02: //HOLD
                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
			event->called_msisdn = channeltab[line].called_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        if(channeltab[line].flag_hold)
                            log->logf("<error> [line %d] Line is already holding, but got HOLD_IND", line);
                        resetLineToIdle(line);
                        channeltab[line].flag_hold = mymsg.adr.adrPLCI;
			/* TODO: we could now wait for DISCONNECT_B3 (?) */
                        return 1;
                        break;
                    case 0x03: { //RETRIEVE

                        // send CONNECT_B3_REQ
                        CONNECT_B3_REQ_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
                        CONNECT_B3_REQ_PLCI(&resp) = mymsg.adr.adrPLCI;
                        CONNECT_B3_REQ_NCPI(&resp) = mymsg.NCPI;
                        sendCapiMessage(line, &resp);
			/* TODO: we might get a 0x04 0x82 (DISCONNECT_IND) here if the caller hangs up
			         in this very moment */
			_cmsg e = waitForEvent(line, 0x82, 0x81, 1);
			freeMessage(line, &e);
			if(!waitForACTIVEIND(line)) 
			    return 0; //FIXME
			channeltab[line].flag_b3 = 1;

                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
			event->called_msisdn = channeltab[line].called_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        channeltab[line].flag_hold = 0;
                        // the state "connected" will be set by CONNECT_B3_CONF, anyway
                        channeltab[line].state = state_connected;
			channeltab[line].flag_b3 = 1;
                        return 1;
                        break;
		    }
                    case 0x07: //3PTY
                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
			event->called_msisdn = channeltab[line].called_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        channeltab[line].flag_3pty = 1;
                        channeltab[line].state = state_connected;
			channeltab[line].flag_b3 = 1;
                        return 1;
                    case 0x08: //3PTY END
			channeltab[line].flag_3pty = 0;
                        break;
                    }
            }
        }
    } else if (IS_DISCONNECT_B3_CONF(&mymsg)) {
	channeltab[line].flag_b3 = 0;
        log->logf("<verbose> [line %d] got disconnect b3 conf", line);
    } else if (IS_DISCONNECT_CONF(&mymsg)) {
        log->logf("<verbose> [line %d] got disconnect conf", line);
    } else if (IS_DISCONNECT_B3_IND(&mymsg)) {
        _cmsg resp;
        DISCONNECT_B3_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&mymsg);
        this->sendCapiMessage(line, &resp);
	channeltab[line].flag_b3 = 0;
        if(STATE(line) == state_connected || STATE(line) == state_holdreq) {
	    struct job*j = &channeltab[line].recordjob;
            STATE(line) = state_disconnecting;
	    if(j->callback) {
		j->callback(j->callback_data, j->data_size);
		j->callback = 0;
	    }
	    lockMutex(&channeltab[line].audio_mutex);
            if(channeltab[line].playjob.bytes_remaining)
                free(channeltab[line].playjob.data);
            channeltab[line].playjob.bytes_remaining = 0;
            channeltab[line].recordjob.bytes_remaining = 0;
	    unlockMutex(&channeltab[line].audio_mutex);
        }
	int reason = DISCONNECT_B3_IND_REASON_B3(&mymsg);
        log->logf ("<verbose> [line %d] b3 reason: %04x", line, reason);
        translatereason(line, reason);
    } else if (IS_DISCONNECT_IND(&mymsg)) {
	/* we always send a response, even if the message is really for
	   some other application. I think Capi expects us to. -mk */
        _cmsg resp;
        DISCONNECT_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&mymsg);
        this->sendCapiMessage(line, &resp);


	/* we have to make sure it's for us */
	if((DISCONNECT_IND_PLCI(&mymsg)&0xff00) != // FIXME: ff00?
	    (channeltab[line].plci_ncci&0xff00)) {

	    if(channeltab[line].flag_hold && 
	       (DISCONNECT_IND_PLCI(&mymsg)&0xff00) ==
	       (channeltab[line].flag_hold&0xff00)) {
		/* call on hold got terminated */
		channeltab[line].flag_hold = 0;
		log->logf("<debug> [line %d] PLCI %08x no longer holding", line,
			DISCONNECT_IND_PLCI(&mymsg));
	    }
	    if(STATE(line) != state_idle) {
		log->logf("<debug> [line %d] ignoring DISCONNECT_IND (%08x!=%08x)", line,
		    DISCONNECT_IND_PLCI(&mymsg), channeltab[line].plci_ncci
		);
		return 0; /* not our plci - these rampant disconnect messages
			     are *very* annoying */
	    }
	}

        int reason = DISCONNECT_IND_REASON(&mymsg);
        if(STATE (line) == state_disconnecting) {
            log->logf ("<notice> [line %d] returning HANGUP Event (terminating a connection)", line);
            log->logf ("<verbose> [line %d] hangup reason: %04x", line, reason);
            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
	    event->called_msisdn = channeltab[line].called_msisdn;
            event->jobid = -1;
            event->tone = -1;
            event->reason = translatereason(line, reason);

	    if(channeltab[line].flag_hold &&
	       channeltab[line].flag_3pty) {
		/* TODO: send DISCONNECT_REQ for second line here */
	    }

            resetLineToIdle(line);
            /*return*/
            return 1;
        } else if(STATE(line) == state_calling) {
            log->logf("<notice> [line %d] returning HANGUP Event (terminating a call job)", line);
            log->logf("<verbose> [line %d] reason: %04x", line, reason);

            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
	    event->called_msisdn = channeltab[line].called_msisdn;
            event->jobid = channeltab[line].callingjob;
            event->tone = -1;
            event->reason = translatereason(line, reason);

            resetLineToIdle(line);
            /*return*/
            return 1;
        } else if(STATE(line) == state_retrievereq ||
                  STATE(line) == state_holdreq ||
                  STATE(line) == state_3ptyreq) {
            log->logf("<notice> [line %d] returning HANGUP Event (terminating a suplementary job)", line);
            log->logf("<verbose> [line %d] reason: %04x", line, reason);

            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
	    event->called_msisdn = channeltab[line].called_msisdn;
            event->jobid = channeltab[line].supplementaryjob;
            event->tone = -1;
            event->reason = translatereason(line, reason);

            resetLineToIdle(line);
            /*return*/
            return 1;
        } else {
            resetLineToIdle(line);
        }
    }
    else if (IS_DATA_B3_CONF(&mymsg)) {
	/* this data confirmation packet has been deliberately let through
	   by the worker thread to indicate that _all_ data has been
	   transferred. We now report this fact back to the application */
	event->line = line;
	event->type = CAPI_EVENT_DONE;
	event->msisdn = channeltab[line].other_msisdn;
	event->called_msisdn = channeltab[line].called_msisdn;
	event->reason = 0;
	event->jobid = channeltab[line].playjob.jobid;
	event->tone = -1;
	/*return*/
	return 1;
    }
    else if(IS_DATA_B3_IND(&mymsg)) {
	event->line = line;
	event->type = CAPI_EVENT_DONE;
	event->msisdn = channeltab[line].other_msisdn;
	event->called_msisdn = channeltab[line].called_msisdn;
	event->reason = 0;
	event->jobid = channeltab[line].recordjob.jobid;
	event->tone = -1;
	struct job*j = &channeltab[line].recordjob;
	if(j->callback) {
	    j->callback(j->callback_data, j->data_size);
	    j->callback = 0;
	}
	/*return*/
	return 1;
    }
    /*else if (IS_MANUFACTURER_IND(&mymsg)) {
       _cmsg resp;
       // send MANUFACTURER_RESP
       MANUFACTURER_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
       MANUFACTURER_RESP_CONTROLLER(&resp) = MANUFACTURER_IND_CONTROLLER(&mymsg);
       MANUFACTURER_RESP_MANUID(&resp) = MANUFACTURER_IND_MANUID(&mymsg);
       MANUFACTURER_RESP_MANUDATA(&resp) = MANUFACTURER_IND_MANUDATA(&mymsg);
       this->sendCapiMessage(line, &resp);
}*/
    else if (IS_RESET_B3_IND(&mymsg)) {
        /* Network Control Connection Identifier */
        int ncci = RESET_B3_IND_NCCI(&mymsg);
        unsigned char*ncpi = RESET_B3_IND_NCPI(&mymsg);
        log->logf("<verbose> [line %d] got RESET_B3_IND: %x %x", line,ncci,ncpi);
        if(ncpi && *ncpi) {
            printf("ncpi length: %d\n", *ncpi);
            int t;
            for(t=1;t<*ncpi;t++) {
                printf("%02x ", ncpi[t]);
            }
        }
        _cmsg resp;
        RESET_B3_RESP_HEADER(&resp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        RESET_B3_RESP_NCCI(&resp) = RESET_B3_RESP_NCCI(&mymsg);
        this->sendCapiMessage(line, &resp);
    } else if (IS_INFO_IND(&mymsg)) {
        /* we get these sometimes when we answer a call.
        * capi: is there something you want to tell us?
	*/
	
        int infonumber = INFO_IND_INFONUMBER(&mymsg);
        log->logf("<verbose> [line %d] INFO_IND: infonumber:%02x", line,infonumber);
        if((infonumber&0xc000)==0) // non-supplementary information element
        {
            //we could now decipher this, according to ETS 300 102-1 Q.931.
            //we're just not in the mood.
        }

	/* These INFO_IND packets also fly by when some other application is answering
	   calls- be careful in doing too much here. */
        _cmsg CMSGresp;
        // send INFO_RESP
        INFO_RESP_HEADER(&CMSGresp, mymsg.ApplId, mymsg.Messagenumber, CONTROLLER(line));
        INFO_RESP_PLCI(&CMSGresp) = INFO_IND_PLCI(&mymsg);
        this->sendCapiMessage(line, &CMSGresp);
    } else {
        log->logf("<error> [line %d] (MessageLoop) unknown message command=0x%02x subcommand=0x%02x", line, mymsg.Command, mymsg.Subcommand);
    }
    return 0;
}

/* This routine reads messages from the controlMessages queue
   (filled in by the worker thread) and processes them, returning
   an event as soon as something interesting has happened */
ISDNEvent CapiHandler2::listen (int line, int&timeout) 
{
    log->logf("<verbose> [line %d] entering Message Loop...", line);
    int handletimeout = 0;
    if (timeout>0)
	handletimeout = 1;

    while(1) {
	_cmsg msg;
	int time = 1000;
	msg = getMessage(line, time);
	if(time) {
            ISDNEvent e;
            e.jobid = -1;
            e.tone = -1;
            if(this->handleMessage(line, msg, &e)) {
		char*eventstr = eventToString(&e);
		log->logf("<verbose> [line %d] listen returns %s", line, eventstr);
		free(eventstr);
		freeMessage(line, &msg);
                return e;
	    }
	    freeMessage(line, &msg);
	}
	if(timeout<=0 && handletimeout) {
	    break; // timeout
	}
	timeout -= 1;
    }

    ISDNEvent e;
    e.type = TIMEOUT;
    e.line = line;
    e.msisdn = 0;
    e.called_msisdn = 0;
    e.reason = 0;
    e.jobid = -1;
    e.tone = -1;

    char*eventstr = eventToString(&e);
    log->logf("<verbose> [line %d] listen returns %s", line, eventstr);
    free(eventstr);
    return e;
}

/* tries to forward the given line to the given msisdn. This still
 * doesn't work. We always get a "network did not answer in time"
 * message.
 */
int CapiHandler2::forward (int line, char*msisdn) {
    /* this won't work with the windows driver of the
     * fritz card, which supports only supplementary services 1-4.
     */
    printf("STATE(line):%d state_idle:%d\n", STATE(line), state_idle);
    if(STATE(line) == state_alerted) // not connected
    {
        _cmsg req;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        // CF word 0x9 =CF
        //    dword handle
        //    word 0x0000 (unconditional)
        //    word basic service (ETS 300 196 (D.5))
        //    struct server user number (facility party number)
        //    struct forwarded to number (facility party number)
        //    struct forwarded to subaddress (called party subaddress)
        // CD word 0xD =CD
        //    word presentation allowed (0=no, 1=yes)
        //    struct deflected to number
        //    struct deflected to subaddress
        //
        //                         len CF     len handle   cond.    service  usernr            forward to         subaddress
        //unsigned char cBuffer[] = {20, 0x9,0, 17, 1,2,3,4, 0,0,     17,1,     3,0x81,0x33,0x33, 3,0x80,0x32,0x39,  0};
        //                        len CD     len pres.  number
        unsigned char cBuffer[] = {10, 0xd,0,  7, 1,0,   3,0x80,0x22,0x21,  0}; //CD - works ok
        //unsigned char cBuffer[] = { 3, 0x0,0, 0}; //INFO - works ok
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->sendCapiMessage(line, &req);
    }
    return 0;
}

/* sends a DTMF */
int CapiHandler2::beep(int line, char*digits)
{
    _cmsg req,conf;
    FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)1; // 1: DTMF
    /*len  send   msdigit mspause tosend  characters */
    int l = strlen(digits);
    unsigned char cBuffer[] = {8,
				   3,0,     //function:send
				   0,1,    //msdigit:256
				   0,1,    //mspause:256
				   0,       //tosend
				   0,       //characteristics
			       0};
    unsigned char*buffer2 = (unsigned char*)malloc(l+10);
    memcpy(buffer2, cBuffer, sizeof(cBuffer));
    buffer2[7] = l;
    strcpy((char*)&buffer2[8], (char*)digits);
    buffer2[8+l] = 0; //characteristics
    buffer2[0] = 1+6+(1+l)+1;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = buffer2;
    sendCapiMessage(line, &req);

    conf = waitForEvent(line,0x80,0x81,1);

    if(FACILITY_CONF_FACILITYSELECTOR(&conf)==1) {
	unsigned char*data = FACILITY_CONF_FACILITYCONFIRMATIONPARAMETER(&conf);
	if(!data || (data[0]<2 || (data[1]!=0 || data[2]!=0))) {
	    log->logf("<error> [line %d] DTMF sending failed.", line);
	}
    } else {
	log->logf("<error> [line %d] wrong selector:%d", line, FACILITY_CONF_FACILITYSELECTOR(&conf));
    }
    freeMessage(line, &conf);
    return 0;
}

/* puts a call on hold. The call won't be actually on hold
 * until a FACILITY_IND message is received
 */
int CapiHandler2::hold(int line) {
    if(!channeltab[line].flag_hold) {
        _cmsg req;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        unsigned char cBuffer[] = {3, 0x2,0, 0}; // HOLD
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->sendCapiMessage(line, &req);
        _cmsg e = this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
	freeMessage(line, &e);
        channeltab[line].supplementaryjob = global_job_counter++;
        channeltab[line].state = state_holdreq;
        return channeltab[line].supplementaryjob;
    } else {
        log->logf("<error> [line %d] Can't put two lines on hold", line);
    }
    return 0;
}

/* tries to retrieve a call. The call won't be actually be retrieved
 * until a FACILITY_IND message is received.
 */
int CapiHandler2::retrieve(int line) {
    if(channeltab[line].flag_hold && channeltab[line].state==state_idle) {
        _cmsg req;
	channeltab[line].plci_ncci = channeltab[line].flag_hold;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        unsigned char cBuffer[] = {3, 0x3,0, 0}; // RETRIEVE
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->sendCapiMessage(line, &req);
        _cmsg e = this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
	freeMessage(line, &e);
        channeltab[line].supplementaryjob = global_job_counter++;
        channeltab[line].state = state_retrievereq;
        return channeltab[line].supplementaryjob;
    } else {
	if(!channeltab[line].flag_hold)
	    log->logf("<warning> [line %d] Can't do retrieve(). No call waiting.", line);
	else
	    log->logf("<warning> [line %d] Can't do retrieve(). Line not idle.", line);
    }
    return 0;
}

int CapiHandler2::endconference(int line) {
    if(!channeltab[line].flag_3pty) {
        log->logf("3 party conference not active.");
        return 0;
    }
    _cmsg req;
    FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
    //                         len ID     len Call
    unsigned char cBuffer[] = {7,  0x8,0, 4,  1,1,1,0}; // 3PTY
    // insert the plci of the line on hold into the struct
    *(int*)&cBuffer[4]=channeltab[line].flag_hold;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
    this->sendCapiMessage(line, &req);

    return 1;
}


/* tries to open up a 3pty with one held and one open call
 * at the moment, 3pty works, but may cause problems with
 * "unexpected" hangups, as we'll retrieve an hangup event
 * for both lines.
 */
int CapiHandler2::conference(int line) {
    if(!channeltab[line].flag_hold) {
        log->logf("There's no line on hold. Can't do three party conference.");
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("There's no connection open. Can't do three party conference.");
        return 0;
    }

    _cmsg req;
    FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
    //                         len ID     len Call
    //unsigned char cBuffer[] = {7,  0x7,0, 4,  1,2,1,0}; // 3PTY
    unsigned char cBuffer[] = {7,  0x7,0, 4,  1,1,1,0}; // 3PTY
    // insert the plci of the line on hold into the struct
    *(int*)&cBuffer[4]=channeltab[line].flag_hold;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
    this->sendCapiMessage(line, &req);
    _cmsg e = this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
    freeMessage(line, &e);
    channeltab[line].supplementaryjob = global_job_counter++;
    channeltab[line].state = state_3ptyreq;
    return channeltab[line].supplementaryjob;
}

/* tries to perform ECT (explicit call transfer)
 * suffers from the same problem as CF
 */
int CapiHandler2::ect(int line) {
    if(!channeltab[line].flag_hold) {
        log->logf("There's no line on hold. Can't do ect.");
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("There's no connection open. Can't do ect.");
        return 0;
    }
    _cmsg msg, req, resp;
    DISCONNECT_B3_REQ_HEADER(&req, APPL_ID(line), 1234, CONTROLLER(line));
    DISCONNECT_B3_REQ_NCCI(&req) = channeltab[line].plci_ncci;
    DISCONNECT_B3_REQ_NCPI(&req) = NULL;
    sendCapiMessage(line, &req);
    
    _cmsg e = this->waitForEvent(line,CAPI_DISCONNECT_B3, CAPI_CONF, 0);
    freeMessage(line, &e);

    FACILITY_REQ_HEADER(&req, APPL_ID(line), (short int)(int)this, CONTROLLER(line));
    FACILITY_REQ_PLCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_NCCI(&req) = channeltab[line].plci_ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
    //                         len ID     len Call
    unsigned char cBuffer[] = {7,  0x6,0, 4,  1,1,1,0}; // 3PTY
    // insert the plci of the line on hold into the struct
    *(int*)&cBuffer[4]=channeltab[line].flag_hold;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
    this->sendCapiMessage(line, &req);

    int num = 0;
    while(1) {
        int timeout = 60*1000;
        msg = getMessage(line, timeout);
        if(timeout <= 0) {
            return 0;
	}
        if(IS_DISCONNECT_B3_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect b3 confirmed.", line);
        } else if(IS_DISCONNECT_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect confirmed.", line);
        } else if(IS_FACILITY_IND(&msg)) {
            ISDNEvent e;
            if(!this->handleMessage(line, msg, &e)) {
		freeMessage(line, &msg);
                return 0; // don't reset line, as the disconnects won't happen
	    }
        }
        else if(IS_DISCONNECT_B3_IND(&msg)) {
            log->logf("<verbose> [line %d] b3 disconnected", line);
            DISCONNECT_B3_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&msg);
            sendCapiMessage(line, &resp);
        } else if(IS_DISCONNECT_IND(&msg)) {
            log->logf("<notice> [line %d] disconnected", line);
            DISCONNECT_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&msg);
            sendCapiMessage(line, &resp);
            num ++;
            if(num == 2) {// both lines will be released
		freeMessage(line, &msg);
                break;
	    }
        }
	freeMessage(line, &msg);
    }
    resetLineToIdle(line);
    return 1;
}

int CapiHandler2::waitForACTIVEIND(int line)
{
    _cmsg resp, msg;
    // wait for CONNECT_B3_ACTIVE_IND
    msg = waitForEvent(line, 0 /*CAPI_CONNECT_B3_ACTIVE*/, CAPI_IND, 1);
    if(!IS_CONNECT_B3_ACTIVE_IND(&msg)) {
	if(IS_DISCONNECT_IND(&msg)) {
	    ISDNEvent e;
	    handleMessage(line, msg, &e);
	    freeMessage(line, &msg);
	} else if(IS_DISCONNECT_B3_IND(&msg)) {
	    ISDNEvent e;
	    handleMessage(line, msg, &e);
	    freeMessage(line, &msg);
	    /* we're now going into the messageloop for some seconds, because
	     * there may be some disconnect_ind about to come up. we don't
	     * want it to pop up anywhere later, when it will maybe confuse
	     * some other routine.
	     */
	    int timeout = 2;
	    this->listen(line,timeout);
	} else {
            channeltab[line].delayedMessages->put(&msg);
	}
	log->logf("<notice> [line %d] waited for CONNECT_B3_ACTIVE, but got %02x-%02x", line, msg.Command, msg.Subcommand);
	return 0;
    }
    channeltab[line].plci_ncci = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
    log->logf("<verbose> [line %d] Setting PLCI to %x", line,channeltab[line].plci_ncci);
    channeltab[line].flag_b3 = 1;

    // send CONNECT_B3_RESP
    CONNECT_B3_ACTIVE_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
    CONNECT_B3_ACTIVE_RESP_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
    sendCapiMessage(line, &resp);
    freeMessage(line, &msg);
    return 1;
}

/* Tries to reject a waiting call. This routine doesn't
 * take the line as parameter, but rather the INCOMING event
 * which is associated with the call to answer. The line
 * variable is extracted from that struct. Also, the internal
 * data of ISDNEvent (which, in this case, is the _cmsg of
 * the CONNECT_IND event) is used.
 */
void CapiHandler2::reject (int line, ISDNEvent *event, int code) 
{
    _cmsg*msg=(_cmsg*)event->internal;
    
    _cmsg mycmsg;

    CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), msg->Messagenumber, CONTROLLER(line));
    CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(msg);

    if(!code)
	code = 1;

    CONNECT_RESP_REJECT(&mycmsg) = code;  // e.g. 2: reject call, normal call clearing

    CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
    CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
    CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;

    sendCapiMessage(line, &mycmsg);
}

/* Tries to answer a waiting call. Like reject(), this function
 * extracts all the data it needs from the event.
 */
int CapiHandler2::answer (int line, ISDNEvent *event) {
    int ret = 0;
    _cmsg tmp;

    // pCMSG should be the CONNECT_IND message
    _cmsg*pCMSG=(_cmsg*)event->internal;
    tmp = *pCMSG;
    free(event->internal);
    pCMSG = &tmp;

    if (STATE(line) == state_alerted) {
        log->logf ("<notice> [line %d] answering call on line %d", line, event->line);

        _cmsg mycmsg;

        CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), pCMSG->Messagenumber, CONTROLLER(line));
        CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(pCMSG);

        CONNECT_RESP_REJECT(&mycmsg) = 0;  // 0: accept call
        //CONNECT_RESP_REJECT(&mycmsg) = 2;  // 2: reject call, normal call clearing

        CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
        CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
        CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;

        sendCapiMessage(line, &mycmsg);
        if (ret) {
            log->logf("<fatal> [line %d] answer:Could not send CONNECT_RESP message.", line);
            return 0;
        }

        // we assume that the responses come all and in exactly this
        // order. If that is not the case, we'd need something more
        // sophisticated.
        _cmsg resp, msg;

        // wait for CONNECT_ACTIVE_IND
        msg = waitForEvent(line, 0 /*CAPI_CONNECT_ACTIVE*/, CAPI_IND, 1);
        if(!IS_CONNECT_ACTIVE_IND(&msg)) {
            if(IS_DISCONNECT_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
		freeMessage(line, &msg);
            } else
		channeltab[line].delayedMessages->put(&msg);
            log->logf("<notice> [line %d] waited for CONNECT_ACTIVE, but got %02x-%02x", line, msg.Command, msg.Subcommand);
            return 0;
        }
	freeMessage(line, &msg);

        // send CONNECT_ACTIVE_RESP
        CONNECT_ACTIVE_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        CONNECT_ACTIVE_RESP_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&msg);
        sendCapiMessage(line, &resp);

        // wait for CONNECT_B3_IND
        msg = waitForEvent(line, 0 /*CAPI_CONNECT_B3*/, CAPI_IND, 1);
        if(!IS_CONNECT_B3_IND(&msg)) {
            if(IS_DISCONNECT_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
		freeMessage(line, &msg);
            } else
		channeltab[line].delayedMessages->put(&msg);
            log->logf("<notice> [line %d] waited for CONNECT_B3, but got %02x-%02x", line, msg.Command, msg.Subcommand);
            return 0;
        }
	freeMessage(line, &msg);

        // send CONNECT_B3_RESP
        CONNECT_B3_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_RESP_NCCI(&resp) = CONNECT_B3_IND_NCCI(&msg);
        CONNECT_B3_RESP_NCPI(&resp) = CONNECT_B3_IND_NCPI(&msg);
        sendCapiMessage(line, &resp);
	channeltab[line].flag_b3 = 1;
        
	if(!waitForACTIVEIND(line)) {
	    return 0;
	}

        // try to enable dial tones

        log->logf("<verbose> [line %d] enabling dial tones...", line);
        // send FACILITY_REQ
        FACILITY_REQ_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        FACILITY_REQ_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
        FACILITY_REQ_FACILITYSELECTOR(&resp) = (unsigned short)1; /* 1: DTMF */
        /*len  listen  msdigit mspause tosend  characters */
        unsigned char cBuffer[] = {8,  1,0,    40,0,   40,0,   0,      0,         0};
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&resp) = cBuffer;
        sendCapiMessage(line, &resp);
        
        // wait for CAPI_FACILITY_CONF
        msg = waitForEvent(line, CAPI_FACILITY, CAPI_CONF, 1);

        if (FACILITY_CONF_FACILITYSELECTOR(&msg) == 1) {  // facilityselector = 1 means DTMF supported
            FLAG_DTMF(line) = 1;
        } else
            log->logf("<error> [line %d] unknown facility_conf: %d (dialtones not supported?)", line,
                      FACILITY_CONF_FACILITYSELECTOR(&mycmsg));

        STATE(line) = state_connected;
	freeMessage(line, &msg);
        return 1;
    } else {
        log->logf ("<error> [line %d] state=%d but though answer was called", line, STATE(line));
    }
    return 0;
}

/* disconnects the given line. It will only disconnect fully connected lines.
 * Also, it will reset the line_t struct.
 */
void CapiHandler2::hangup (int line) {
    if (STATE(line) != state_connected &&
            STATE(line) != state_holdreq &&
            STATE(line) != state_3ptyreq &&
	    !channeltab[line].flag_hold) {
        log->logf("<error> [line %d] hangup: can't hangup - no connection open", line);
        return;
    }
    // b3 req
    _cmsg msg, resp;
    if(channeltab[line].flag_b3)
    {
	DISCONNECT_B3_REQ_HEADER(&msg, APPL_ID(line), 1234, CONTROLLER(line));
	DISCONNECT_B3_REQ_NCCI(&msg) = channeltab[line].plci_ncci;
	DISCONNECT_B3_REQ_NCPI(&msg) = NULL;
	sendCapiMessage(line, &msg);
    }

    DISCONNECT_REQ_HEADER(&msg, APPL_ID(line), 1235, CONTROLLER(line));
    DISCONNECT_REQ_PLCI(&msg) = channeltab[line].plci_ncci;
    sendCapiMessage(line, &msg);

    while(1) {
        int timeout = 60*1000;
        msg = getMessage(line, timeout);
        if(timeout <= 0)
            break;
        if(IS_DISCONNECT_B3_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect b3 confirmed.", line);
        } else if(IS_DISCONNECT_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect confirmed.", line);
        } else if(IS_DISCONNECT_B3_IND(&msg)) {
	    channeltab[line].flag_b3 = 0;
            log->logf("<verbose> [line %d] b3 disconnected", line);
            DISCONNECT_B3_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&msg);
            sendCapiMessage(line, &resp);
        } else if(IS_DISCONNECT_IND(&msg)) {
	    channeltab[line].flag_b3 = 0;
            log->logf("<notice> [line %d] disconnected", line);
            DISCONNECT_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&msg);
            sendCapiMessage(line, &resp);
	    freeMessage(line, &msg);
            break;
        }
	freeMessage(line, &msg);
    }

    resetLineToIdle(line);
}

/* convert an event into readable form. does allocate a
 * string, which you should free afterwards!
 */
char* CapiHandler2::eventToString (ISDNEvent*event) {
    char buf[256];
    char*name;
    switch(event->type) {
    case TIMEOUT:
        name="TIMEOUT";
        break;
    case CAPI_EVENT_HANGUP:
        name="HANGUP";
        break;
    case CAPI_EVENT_INCOMING:
        name="INCOMING";
        break;
    case CAPI_EVENT_DONE:
        name="DONE";
        break;
    case CAPI_EVENT_DIALTONE:
        name="DIALTONE";
        break;
    case CAPI_EVENT_CONNECTED:
        name="CONNECTED";
        break;
    }
    int s=sprintf(buf, "EVENT [%d] name=%s msisdn=\"%s\" to_misdn=\"%s\"", event->line, name, event->msisdn, event->called_msisdn);
    if (event->type == CAPI_EVENT_DIALTONE)
        sprintf(&buf[s], " tone='%c'", event->tone);
    if (event->type == CAPI_EVENT_DONE)
        sprintf(&buf[s], " job=%d", event->jobid);
    if (event->type == CAPI_EVENT_HANGUP && event->reason)
        sprintf(&buf[s], " reason=%d", event->reason);
    return strdup(buf);
}

	
void CapiHandler2::installInputHandler(int line, InputHandler*handler)
{
    log->logf("[line %d] installInputHandler", line);

    lockMutex(&channeltab[line].audio_mutex);
    channeltab[line].inputHandler = handler;
    unlockMutex(&channeltab[line].audio_mutex);
}
	
void CapiHandler2::installOutputHandler(int line, OutputHandler*handler)
{
    log->logf("[line %d] installOutputHandler", line);

    if(channeltab[line].playjob.bytes_remaining)
	// may fail if the job has been completed meanwhile.
	this->stop(line, channeltab[line].playjob.jobid);

    lockMutex(&channeltab[line].audio_mutex);
    channeltab[line].outputHandler = handler;

    if(STATE(line) == state_connected) {
	int t;
	for(t=0;t<CONNECT_READ_FORWARD;t++)
	    transferData(line);
    }
    unlockMutex(&channeltab[line].audio_mutex);
}

