/* SimpleCapi.cpp

implementation of the SimpleCapi class. (See SimpleCapi.h)

Part of the caiviar package. 

Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include "../config.h"
#if defined(LINUX) || defined(CYGWIN)
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "CapiHandler2.h"
#include "Log.h"
#include "SimpleCapi.h"
#include "wav.h"
#elif defined(WIN32)
#include "CapiHandler2.h"
#include "Log.h"
#include "SimpleCapi.h"
#include <conio.h>
#include "wav.h"
#endif
#include "os.h"

#define CAPI ((CapiHandler2*)(this->internal))
#define LOG ((((CapiHandler2*)(this->internal))->log))

Mutex line_mutex[64];

int instancecount = 0;

static char* properties_msisdn = 0;
static char* properties_controller = 0; // all controllers

SimpleCapi::SimpleCapi(Logger*log) 
{
    int extern_logger = 1;
   
    if(!log) {
	log = new Logger();
	log->initLog (0,-1,0,0,-1,2); //-1)?
	extern_logger = 0;
    }

    this->internal = (void*)(new CapiHandler2(log));
    ((CapiHandler2*)(this->internal))->extern_logger = extern_logger;
    ((CapiHandler2*)(this->internal))->allcalls_hack = 0;

    instancecount++;
    if(instancecount>=2) {
        LOG->logf("<warning> Only one instance of SimpleCapi should be created.");
    }
}

void SimpleCapi::setProperty(char*name, char*value)
{
    if(!name || !value)
	return;
    LOG->logf("<verbose> Setting SimpleCapi property \"%s\"=\"%s\"", name, value);
    if (!strcasecmp(name, "msisdn")) {
        properties_msisdn=strdup(value);
    }
    else if (!strcasecmp(name, "controller")) {
        properties_controller=strdup(value);
    }
    else if (!strcasecmp(name, "allcalls")) {
        int properties_allcalls=atoi(value);
	((CapiHandler2*)(this->internal))->allcalls_hack = properties_allcalls;
    }
}

SimpleCapi::~SimpleCapi() 
{
    LOG->log("<verbose> SimpleCapi::~SimpleCapi()");
    Logger*log = LOG;
    if(!(((CapiHandler2*)(this->internal))->extern_logger))
	delete log;
    delete CAPI;
    instancecount--;
}

int SimpleCapi::open (char*msisdn) 
{
    LOG->logf("<verbose> SimpleCapi::open(%s)",msisdn);
    if(!msisdn || !*msisdn) {
        msisdn = properties_msisdn;
	LOG->logf("<verbose> SimpleCapi::open(): using msisdn \"%s\" and controllermask [%s] from properties", msisdn,
		properties_controller?properties_controller:"all");
    }
    lockMutex(&CAPI->capi_mutex);
    int line = CAPI->open(msisdn, properties_controller);
    unlockMutex(&CAPI->capi_mutex);
    if(line>=0)
	initMutex(&line_mutex[line]);
    LOG->logf("<verbose> [line %d] SimpleCapi::open() returns line %d", line, line);
    return line;
}

void SimpleCapi::close (int line) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::close(%d)", line, line);
    if(!CAPI->checkLine(line))
        return;
    lockMutex(&line_mutex[line]);
    CAPI->close(line);
    unlockMutex(&line_mutex[line]);

    destroyMutex(&line_mutex[line]);
}

void SimpleCapi::acceptCalls (int line) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::acceptCalls(%d)", line, line);
    if(!CAPI->checkLine(line))
        return;
    lockMutex(&line_mutex[line]);
    CAPI->acceptCalls(line);
    unlockMutex(&line_mutex[line]);
}

int SimpleCapi::call (int line, char* msisdn, int timeout) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::call(%d, %s, %d)",line, line,msisdn,timeout);
    if(!CAPI->checkLine(line))
        return 0;
    lockMutex(&line_mutex[line]);
    int callevent = CAPI->call (line, msisdn, -1);
    unlockMutex(&line_mutex[line]);

    int timeout2=timeout;
    while(1) {
        lockMutex(&line_mutex[line]);
        ISDNEvent event = CAPI->listen(line, timeout2);
        unlockMutex(&line_mutex[line]);
        if(event.type == CAPI_EVENT_CONNECTED) {
	    break;
        }
        /* CapiHandler2 will return an HANGUP event, even
         * if there hasn't been a connection yet. this means
         * for some reason we didn't manage to call the other
         * party
         */
        if(event.type == CAPI_EVENT_HANGUP) {
            if(event.reason)
                return event.reason;
            else
                return CAPI_REASON_TECHNICAL;
        }
        if(event.type == TIMEOUT) {
	    stop(line, callevent);
            return CAPI_REASON_NOANSWER;
        }
    }
    return 0;
}

int SimpleCapi::call_nonblock (int line, char* msisdn) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::call_nonblock(%d, %s)",line, line, msisdn);
    if(!CAPI->checkLine(line))
        return -1;
    lockMutex(&line_mutex[line]);
    int retval  = CAPI->call (line, msisdn, -1);
    unlockMutex(&line_mutex[line]);

    return retval;
}

int SimpleCapi::play (int line, const char* filename) 
{
    FILE*fi = fopen (filename,"rb");
    if(!fi) {
        LOG->logf("<error> [line %d] Could not find file\"%s\"!", line, filename);
        return -1;
    }
    fseek (fi, 0, SEEK_END);
    long int size = ftell(fi);
    fseek (fi, 0, SEEK_SET);

    LOG->logf ("<verbose> [line %d] SimpleCapi::play(%d, %s) (%d bytes)", line, line, filename, size);

    char*data = (char*)malloc (size);
    fread (data, 1, size, fi);
    fclose (fi);
    int playjob = play ( line, data, size);
    free (data);
    return playjob;
}

int SimpleCapi::play (int line, void*mem, int len) 
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::play(%d, ptr, %d)", line, line, len);
    int riff = 0;
    unsigned char*mem2 = (unsigned char*)mem;
    if(mem2[0] == 0x52 && mem2[1] == 0x49 && mem2[2] == 0x46 && mem2[3] == 0x46) //RIFF
    {
        LOG->logf("<notice> [line %d] Performing wav to law conversion", line);
        mem = wav2law(mem2, len, &len);
        riff = 1;
    }
    if(!CAPI->checkLine(line))
        return -1;
    lockMutex(&line_mutex[line]);
    int ret=CAPI->play (line, mem, len);
    unlockMutex(&line_mutex[line]);
    if(riff)
        free(mem);
    return ret;
}

int SimpleCapi::record (int line, void*buf, int bufsize) 
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::record(%d, ptr, %d)",line, line,bufsize);
    if(!CAPI->checkLine(line))
        return -1;
    lockMutex(&line_mutex[line]);
    int ret = CAPI->record(line,buf,bufsize,0,0);
    unlockMutex(&line_mutex[line]);

    return ret;
}

struct record_t
{
    void*buf;
    int len;
    char* filename;
    int line;
    Logger*log;
};

void record_end_callback(void*data, int size)
{
    struct record_t*r = (struct record_t*)data;
    struct WAV wav;
    r->log->logf ("<verbose> [line %d] performing law to wav conversion (%d/%d bytes) (\"%s\")", r->line,
	    size,r->len,r->filename);
    convertLAW2WAV(r->buf,size,&wav);
    writeWAV(r->filename,&wav);
    free(wav.data);
    free(r->buf);
    free(r->filename);
    free(data);
}
    
int SimpleCapi::recordWAV (int line, char*filename, int len)
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::recordWAV(%d, %s, %d)", line, line, filename, len);
    if(!CAPI->checkLine(line))
        return -1;
    struct record_t*record =
	(struct record_t*)malloc(sizeof(struct record_t));
    void*buf = malloc(len);
    record->line = line;
    record->buf = buf;
    record->len = len;
    record->filename = strdup(filename);
    record->log = LOG;

    lockMutex(&line_mutex[line]);
    int jobid = CAPI->record(line, buf, len, record_end_callback, record);
    unlockMutex(&line_mutex[line]);

    return jobid;
}

int SimpleCapi::getRecordSize (int line) 
{
    lockMutex(&line_mutex[line]);
    int ret = CAPI->getRecordSize(line);
    unlockMutex(&line_mutex[line]);
    return ret;
}

int SimpleCapi::fax (int line, char*msisdn, char*filename) 
{
    int faxformat = 5; //ascii
    {
        FILE*fi= fopen(filename, "rb");
        char a[4];
        if(fi && fread(a,1,4,fi)==4) {
            if (!strncmp(a,"Sfff",4))
                faxformat = 0;
	    fclose(fi);
        }
    }

    LOG->logf ("<verbose> [line %d] SimpleCapi::fax(%d, %s, %s) (faxformat = %d)", line, line,msisdn,filename, faxformat);
    lockMutex(&line_mutex[line]);
    if(!CAPI->checkLine(line))
        return 0;
    int ret =  CAPI->call (line, msisdn, faxformat);
    unlockMutex(&line_mutex[line]);
    if (!ret) {
        LOG->logf("<verbose> [line %d] no answer\n", line);
        return 0;
    }
    int done=0;
    while (!done) {
        ISDNEvent event = listen(line, 0);
        switch (event.type) {
        case CAPI_EVENT_CONNECTED:
            done = 1;
            break;
        case CAPI_EVENT_HANGUP:
	    return 0;
        default:
	    char*eventstr = eventToString (&event); 
            LOG->logf("<verbose> [line %d] Unexpected: %s\n", line, eventstr);
	    free(eventstr);
            break;
        }
    }

    /* to "play" means to transfer data, which is exactly what we want to
     * do here */
    int playjob = play (line, filename);
    if(playjob<0)
	return 0;
    done=0;
    while (!done) {
        ISDNEvent event = listen(line, 0);
        switch (event.type) {
        case CAPI_EVENT_HANGUP:
            LOG->logf("<verbose> [line %d] hung up\n", line);
            return 0;

        case CAPI_EVENT_DONE:
            if (event.jobid == playjob) {
                done = 1;
            }
            break;

        default:
	    char*eventstr = eventToString (&event); 
            LOG->logf("<verbose> [line %d] Unexpected: %s\n", line, eventstr);
	    free(eventstr);
            break;
        }
    }

    done = 0;
    while(!done) {
        ISDNEvent event = listen(line, 1);
        if(event.type == CAPI_EVENT_HANGUP) {
            done=1;
        }
    }

    return 1;
}

void SimpleCapi::stop (int line, int jobid) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::stop(%d, %d)", line, line, jobid);
    lockMutex(&line_mutex[line]);
    CAPI->stop(line, jobid);
    unlockMutex(&line_mutex[line]);
}

ISDNEvent SimpleCapi::listen (int line, int timeout) 
{
    int timeout2 = timeout;
    LOG->logf("<verbose> [line %d] SimpleCapi::listen(%d, %d)", line, line, timeout2);
    ISDNEvent ret;
    ret.type = -1;
    if(!CAPI->checkLine(line))
        return ret;
    lockMutex(&line_mutex[line]);
    ret = CAPI->listen(line, timeout2);
    unlockMutex(&line_mutex[line]);
    return ret;
}

int SimpleCapi::answer (int line, ISDNEvent*event) 
{
    char*eventstr = eventToString(event);
    LOG->logf("<verbose> [line %d] SimpleCapi::answer(%d, %s)", line, line, eventstr);
    free(eventstr);
    if(!CAPI->checkLine(line))
        return 0;
    lockMutex(&line_mutex[line]);
    int ret = CAPI->answer(line, event);
    unlockMutex(&line_mutex[line]);
    return ret;
}

void SimpleCapi::reject(int line, ISDNEvent*event, int code) 
{
    char*eventstr = eventToString(event);
    LOG->logf("<verbose> [line %d] SimpleCapi::reject(%d, %s)", line, line, eventstr);
    free(eventstr);
    if(!CAPI->checkLine(line))
        return;
    lockMutex(&line_mutex[line]);
    CAPI->reject(line, event, code);
    unlockMutex(&line_mutex[line]);
}


void SimpleCapi::beep (int line, char*dtmf)
{
    int t, l;
    char*s;
    LOG->logf("<verbose> [line %d] SimpleCapi::beep(%d,%s)", line, line, dtmf);
    if(!CAPI->checkLine(line))
        return;
    s = strdup(dtmf);
    l = strlen(s); 
    for(t=0;t<l;t++) {
	/* poor man's toupper() */
	if(s[t]>='a' && s[t]<='d')
	    s[t]+=('A'-'a');
	if(!strchr("0123456789ABCD*#", s[t])) {
	    LOG->logf("<warning> [line %d] not a valid DTMF: %02x %c", line, s[t], s[t]);
	}
    }
    lockMutex(&line_mutex[line]);
    CAPI->beep(line, s);
    unlockMutex(&line_mutex[line]);
    free(s);
}

/* perform several supplemental services. 
   hold and retrieve should now be used through
   SimpleCapi::hold and SimpleCapi::retrieve.
   ect (explicit call transfer) and forwarding
   are untested and probably not working */
int SimpleCapi::supplemental (int line, char* command) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::supplemental(%d, %s)", line, line, command);
    if(!CAPI->checkLine(line))
        return 0;
    lockMutex(&line_mutex[line]);
    int ret;

    if(!strcmp(command, "hold"))
        ret = CAPI->hold(line);
    else if(!strcmp(command, "retrieve"))
        ret = CAPI->retrieve(line);
    else if(!strcmp(command, "3pty"))
        ret = CAPI->conference(line);
    else if(!strcmp(command, "ect"))
        ret = CAPI->ect(line);
    else
        ret = CAPI->forward(line, command);

    unlockMutex(&line_mutex[line]);
    return ret;
}

int SimpleCapi::hold (int line)
{
    LOG->logf("<verbose> [line %d] SimpleCapi::hold(%d)", line, line);
    int jobid;
    jobid = this->supplemental(line, "hold");
    LOG->logf("<verbose> [line %d] SimpleCapi::hold jobid:%d", line, jobid);

    if(!jobid)
	return 0;
    ISDNEvent event;
    while(1)
    {
	event = this->listen(line, 10);
	switch(event.type)
	{
	    case TIMEOUT:
		return 0;
	    case CAPI_EVENT_HANGUP:
	    /* (!) we return 1 here, even though the hold didn't succeed. 
		   however the line is free and it's upon the subsequent 
		   retrieve() to return an error. */
		return 1;
	    case CAPI_EVENT_DONE:
		if(jobid == event.jobid) {
		    /* big hack - as CapiHandler needs some extra time to
		       complete the hold() procedure, we call listen again
		       and hope the event is a TIMEOUT */
		    this->listen(line, 1);
		    return 1;
		}
		LOG->logf("<verbose> [line %d] Warning: job %d done while waiting for hold().", line, event.jobid);
	    break;
	}
    }
}

int SimpleCapi::retrieve (int line)
{
    LOG->logf("<verbose> [line %d] SimpleCapi::hold(%d)", line, line);
    int jobid;
    jobid = this->supplemental(line, "retrieve");
    if(!jobid)
	return 0;
    ISDNEvent event;
    while(1)
    {
	event = this->listen(line, 10);
	switch(event.type)
	{
	    case TIMEOUT:
		return 0;
	    case CAPI_EVENT_HANGUP:
		return 0;
	    case CAPI_EVENT_DONE:
		if(jobid == event.jobid) {
		    /* big hack - as CapiHandler needs some extra time to
		       complete the retrieve() procedure, we call listen again
		       and hope the event is a TIMEOUT */
		    this->listen(line, 1);
		    return 1;
		}
		LOG->logf("<verbose> [line %d] Warning: job %d done while waiting for hold().", line, jobid);
	    break;
	}
    }
}

int SimpleCapi::forward (int line, char*msisdn) 
{
    LOG->logf("<verbose> [line %d] SimpleCapi::forward(%d, %s)", line, line, msisdn);
    LOG->logf("<warning> [line %d] forward() is deprecated. Please use hold(), retrieve() and supplemental().", line);
    return this->supplemental (line, msisdn);
}

void SimpleCapi::hangup (int line) 
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::hangup(%d)", line, line);
    if(!CAPI->checkLine(line))
        return ;
    lockMutex(&line_mutex[line]);
    CAPI->hangup(line);
    unlockMutex(&line_mutex[line]);
}

char* SimpleCapi::eventToString (ISDNEvent*event) 
{
    LOG->logf("<verbose> SimpleCapi::eventToString(line:%d)", event->line);
    char* ret = CAPI->eventToString(event);
    return ret;
}

class LineConnector: public InputHandler, public OutputHandler {
    unsigned char*ringbuffer;
    int readpos;
    int writepos;
    int buffersize;
    Logger*log;

    Mutex mutex;

    public:
    /* handleOutput is expected to fill data[] up to len 
       it should return an integer <=len, conforming to
       the number of bytes written into data[].
     */
    LineConnector(Logger*log);

    //non virtual destructor:
    ~LineConnector();
    virtual int handleOutput(unsigned char*data, int len);
    virtual void handleInput(unsigned char*data, int len);
};

LineConnector::LineConnector(Logger*log)
{
    buffersize = 2048*8*2;
    ringbuffer = new unsigned char[buffersize];
    readpos = 0;
    writepos = 0;
    this->log = log;
    initMutex(&mutex);
}

LineConnector::~LineConnector()
{
    delete ringbuffer;
}

void LineConnector::handleInput(unsigned char*data, int len)
{
    lockMutex(&mutex);
    log->logf("<debug> LineConnector[%08x]::handleInput(%d) readpos=%d",this, len, readpos);
    if( (readpos < writepos && readpos + len >= writepos) ||
	(readpos >  writepos && readpos + len >= buffersize && (readpos + len) % buffersize >= writepos)
      ) {
	/* ignoring */
	log->logf("<debug> ignoring Input writepos=%d readpos=%d len=%d buffersize=%d", writepos, readpos, len, buffersize);
	unlockMutex(&mutex);
	return;
    }
    if(readpos + len > buffersize) {
	int read1 = buffersize-readpos;
	memcpy(&ringbuffer[readpos], data, read1);
	memcpy(&ringbuffer[0], &data[read1], len - read1);
	readpos = len - read1;
    } else {
	memcpy(&ringbuffer[readpos], data, len);
	readpos += len;
	readpos %= buffersize;
    }
    unlockMutex(&mutex);
}

int LineConnector::handleOutput(unsigned char*data, int len)
{
    lockMutex(&mutex);
    log->logf("<debug> LineConnector[%08x]::handleOutput(%d) writepos=%d",this, len, writepos);
    if( (writepos <= readpos && writepos + len > readpos) ||
	(writepos > readpos && writepos + len >= buffersize && (writepos + len) % buffersize > readpos)
      ) {
	/* ignoring */
	log->logf("<debug> ignoring Output writepos=%d readpos=%d len=%d buffersize=%d", writepos, readpos, len, buffersize);
	unlockMutex(&mutex);
	return 0;
    }
    if(writepos + len > buffersize) {
	int read1 = buffersize-writepos;
	memcpy(data, &ringbuffer[writepos], read1);
	memcpy(&data[read1], &ringbuffer[0], len - read1);
	writepos = len - read1;
    } else {
	memcpy(data, &ringbuffer[writepos], len);
	writepos += len;
	writepos %= buffersize;
    }
    unlockMutex(&mutex);
    return len;
}

void SimpleCapi::connect(int line1, int line2)
{
    LineConnector*lc1 = new LineConnector(LOG);
    LineConnector*lc2 = new LineConnector(LOG);
    LOG->logf("<verbose> SimpleCapi::connect(%d, %d)", line1, line2);
    CAPI->installInputHandler(line1, lc1);
    CAPI->installInputHandler(line2, lc2);
    Sleep(50); //wait for some data to come in
    CAPI->installOutputHandler(line2, lc1);
    CAPI->installOutputHandler(line1, lc2);
}

void SimpleCapi::installInputHandler (int line, InputHandler*handler) 
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::installInputHandler(%d)", line, line);
    if(!CAPI->checkLine(line))
        return ;
    lockMutex(&line_mutex[line]);
    CAPI->installInputHandler(line, handler);
    unlockMutex(&line_mutex[line]);
}

void SimpleCapi::installOutputHandler (int line, OutputHandler*handler) 
{
    LOG->logf ("<verbose> [line %d] SimpleCapi::installOutputHandler(%d)", line, line);
    if(!CAPI->checkLine(line))
        return ;
    lockMutex(&line_mutex[line]);
    CAPI->installOutputHandler(line, handler);
    unlockMutex(&line_mutex[line]);
}


