/* SimpleCapi.h

header for SimpleCapi.cpp, prototypes for the SimpleCapi
class, which provides basic ivr functions.

Part of the caiviar package. 

Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#ifndef SimpleCapi_h
#define SimpleCapi_h

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "Log.h"

/* in the documentation of this interface, the terms "line" and
 * "channel" both mean the same.
 */

#define TIMEOUT 0
#define CAPI_EVENT_HANGUP 1    // connection got terminated
#define CAPI_EVENT_INCOMING 2  // incoming call
#define CAPI_EVENT_DONE 3      // job is done
#define CAPI_EVENT_DIALTONE 4  // other end pressed some keys
#define CAPI_EVENT_CONNECTED 5 // an outgoing call got answered

#define CAPI_REASON_TECHNICAL 1
#define CAPI_REASON_BUSY 2
#define CAPI_REASON_WRONGNUMBER 3 //"kein anschluss unter dieser nummer" - in germany
#define CAPI_REASON_REJECTED 4
#define CAPI_REASON_NOANSWER 5

struct ISDNEvent {
    int line;
    int type;
    char*msisdn;
    char*called_msisdn;

    int jobid;  // only set if type==CAPI_EVENT_DONE
    int tone;   // only set if type==CAPI_EVENT_DIALTONE
    int reason; // only set if type==CAPI_EVENT_HANGUP

    void* internal; // for internal use. do not read, write, modify.
}
;

class InputHandler;
class OutputHandler;

class SimpleCapi {
    void*internal;
public:
    /*constructor
     *
     * @param propertyfile The file name of a textfile defining some
     *                     properties needed by capi. (e.g. about logging)
     */
    SimpleCapi(Logger*log);
    /*destructor*/
    ~SimpleCapi();

    /* set's a SimpleCapi specific property. 
       Current properties include: "msisdn", "allcalls", "controller".
       @param name The property name
       @param value The property value
     */
    void setProperty(char*name, char*value);

    /* opens a new isdn channel, and returns an unique id for accessing it.
     *
     * @param msisdn the msisdn of the local station. 
     *
     * @returns the line id
     */
    int open (char * msisdn);

    /* closes an isdn channel, therefore giving it free.
     *
     * @param line  the line id
     */
    void close (int line);

    /* Send DTMFs.
     *
     * @param line  the line id
     * @param dtmf  the dialtones to send, as null-terminated string. Only the characters
                    '0'-'9','#','*' are supported, everything else will cause the dtmf string
		    to be rejected.
     */
    void beep (int line, char*dtmf);

    /*
     * This is the function to use if one want's to listen for calls. If
     * this is called, incoming connections will trigger a CAPI_EVENT_INCOMING
     * event. Note that to reset this, you'll have to close the line and
     * open another.
     * @param line  the line to listen on
     */
    void acceptCalls (int line);

    /* call the specified msisdn.
     * 
     * @param line  the line to use
     * @param msisdn  the msisdn to call
     * @param timeout  the number of seconds to wait for an answer
     * @returns 0  if somebody answered in time, REASON otherwise
     */
    int call (int line, char* msisdn, int timeout);

    /* call the specified msisdn, return immediately. It returns
     * a job id which can be used to wait for the call to 
     * succeed (CAPI_EVENT_DONE) or fail (CAPI_EVENT_HANGUP)
     * @param  line the line to use
     * @param  msisdn the misdn to call
     * @returns  a job id for the calling task
     */
    int call_nonblock (int line, char* msisdn);

    /* send a sound file into a line.  This call does not block,
     * but rather returns a job id which can be used
     * for waiting for the end of the sound file.
     *      @param  line the line on which to play the sound file
     *	    @param  filename the filename of the soundfile
     *	    @param  mem the memory the sound file is in. Capi will copy
     *	            the sound data to an own buffer, so you may free
     *	            this buffer afterwards.
     *	    @param  len the length of the sound file
     *	    @returns  a job id. An event with this job id will be
     *	              generated when the end of the audio data is 
     *	              reached.
     */
    int play (int line, const char* filename);
    int play (int line, void*mem, int len);

    /* record the other end's voice into buf. This call does not block,
     * but rather returns a job id which can be used for waiting for
     * the end of an recording. Notice that other events (especially
     * DIALTONE and HANGUP) will still be sent.
     *	    @param line the line id
     *	    @param buf the buffer to copy the audio data to
     *	    @param bufsize the maximal size of the audio data
     *	    @returns a job id, which can be used to wait for the end
     *	     of the recording
     */
    int record (int line, void*buf, int bufsize);

    /* record the other end's voice into a WAV file. The file will first be recorded,
     * and then converted to a WAV file after a) len samples have been recorded (in which
     * case a DONE event is sent) or b) the line is being disconnected.
     *
     *	    @param line the line id
     *	    @param filename the filename of the resulting WAV file
     *      @param len maximum number of samples to record (=seconds to record/8000)
     *	    @returns a job id, which can be used to wait for the end
     *	     of the recording
     */
    int recordWAV (int line, char*filename, int len);

    /* return the number of bytes read by last record command
     *      @param line the line id
     *      @returns number of bytes read
     */
    int getRecordSize (int line);
    
    /* send a fax through the line. This call blocks till the fax
     * has been transmitted.
     * @param line the line id
     * @param msisdn the msisdn of the fax machine
     * @param filename the filename of the fax file
     */
    int fax (int line, char*msisdn, char*filename);

    /* stop a running job. Most likely you may want to stop jobs
     * which play soundfiles once in a while (for barge-in functionality)
     * @param jobid the job to stop
     */
    void stop (int line, int jobid);

    /* wait for something to happen. This might be a hangup by the other
     * end, a job that finished, an incoming call etc.
     *	    @param line the line to listen on. If -1, listen on every line
     *	    @param timeout the time to listen. If nothing happened while
     *                we were waiting, return a TIMEOUT event. If timeout=0, 
     *                wait forever.
     */
    ISDNEvent listen (int line, int timeout);

    /* answer an incoming call.
     *	    @param event the INCOMING event to answer
     *	    @returns true on success, false on failure (e.g. if the caller
     *                hung up already)
     */
    int answer (int line, ISDNEvent* event);

    /* reject an incoming call.
     *	    @param event the INCOMING event to reject
    *	    @param code sort of reject: 1=ignore, 2=hangup, 3=busy
     */
    void reject (int line, ISDNEvent* event, int code);

    /* put existing call on hold.
     *	    @param line the line to put on hold. must be connected.
     *	    @returns true on success, false on failure (e.g. if the caller
     *                hung up already)
     *
     */
    int hold(int line);
 
    /* retrieve call on hold.
     *	    @param line the line to retrieve. must be on hold.
     *	    @returns true on success, false on failure (e.g. if the caller
     *                hung up already)
     *
     */
    int retrieve(int line);

    /* do a supplemental service command, e.g. "3conf".
     *	    @param line the line to use
     *	    @param command the command to send, e.g. "3conf".
     *	    @returns true on success, false on failure.
     */
    int supplemental (int line, char* command);
   

    int forward (int line, char*msisdn); /* *deprecated* - see supplemental */


    /* hangup on a connection. This call blocks until the connection is correctly
     * terminated.
     * @param line the line id
     */
    void hangup (int line);

    /* convert an event to string. Useful for debugging.
     * @param event the Event to convert
     * @returns  a newly allocated string. the caller has to free it.
     */
    char* eventToString (ISDNEvent*event);

    /* connect the audio streams of two lines. This is similar to 3conf, except
       that the local computer performs the merging of the audio data. (As opposed
       to the telephone provider). You can use record() on connected lines, but not
       play().
     * @param line1 one line
     * @param line1 the other line
     */
    void connect(int line1, int line2);

    /* Installs an input handler, i.e. a class which reads and processes data
     * received on the ISDN bus. (Usually audio data)
     * @param line the line
     * @param handler the InputHandler to install
     */
    void installInputHandler(int line, InputHandler*handler);

    /* Installs an output handler, i.e. a class which generates data to
     * send on the ISDN line. (Usually audio data)
     * @param line the line
     * @param handler the OutputHandler to install
     */
    void installOutputHandler(int line, OutputHandler*handler);
};

class InputHandler {
    public:
    /* handleInput gets len bytes of data to handle */
    virtual void handleInput(unsigned char*data, int len)=0;
};

class OutputHandler {
    public:
    /* handleOutput is expected to fill data[] with at most len
       bytes.  it should return an integer <=len, matching 
       the number of bytes written into data[].
     */
    virtual int handleOutput(unsigned char*data, int len)=0;
};

#endif
