/* wav.c
   Routines for handling .wav files

   Part of the caiviar package.
   
   Copyright (c) 2001 Matthias Kramm <kramm@quiss.org>

   This file is distributed under the GPL, see file COPYING for details */

#include "../config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "wav.h"

struct WAVBlock {
    char id[4];
    unsigned int size;
};

struct reader {
    FILE*fi;
    unsigned char*data;
    int len;
    int pos;
};

int reader_read(void*mem, int size, int num, struct reader*read) {
    if(read->fi)
	return fread(mem,size,num,read->fi);
    else {
	int n = size*num;
	if(read->pos+n > read->len) {
	    n = read->len - read->pos;
	    n/=size; n*=size;
	}
	memcpy(mem, &read->data[read->pos],n);
	read->pos += n;
	return n/size;
    }
}
void reader_goto(struct reader*read, int pos)
{
    if(read->fi) {
	fseek(read->fi, pos, SEEK_SET);
    } else {
	read->pos = pos;
    }
}
int reader_getlen(struct reader*read)
{
    if(read->fi) {
	int filesize;
	fseek(read->fi, 0, SEEK_END);
	filesize = ftell(read->fi);
	fseek(read->fi, 0, SEEK_SET);
	return filesize;
    } else {
	return read->len;
    }
}
int reader_getpos(struct reader*read)
{
    if(read->fi) {
	return ftell(read->fi);
    } else {
	return read->pos;
    }
}

int getWAVBlock(struct reader*fi, struct WAVBlock*block)
{
    unsigned int size;
    unsigned char b[4];
    if(reader_read(block->id,1,4,fi)<4)
	return 0;
    if(reader_read(b,1,4,fi)<4)
	return 0;
    block->size = b[0]|b[1]<<8|b[2]<<16|b[3]<<24;
    /*printf("Chunk: [%c%c%c%c] (%d bytes)\n", 
	    block->id[0],block->id[1],
	    block->id[2],block->id[3],
	    block->size);*/
    return 1;
}


int readWAV2(struct reader*fi, struct WAV*wav)
{
    unsigned char b[16];
    long int filesize;
    struct WAVBlock block;
    long int pos = 0;
    filesize = reader_getlen(fi);
    memset(wav,0,sizeof(struct WAV));

    //printf("Filesize: %d\n", filesize);

    if(!getWAVBlock (fi, &block)) {
	fprintf(stderr, "readWAV: Couldn't get first block\n");
	return 0;
    }
    if(strncmp(block.id,"RIFF",4)) {
	fprintf(stderr, "readWAV: not a wav file\n");
	return 0;
    }
    if(block.size + 8 < filesize)
	fprintf(stderr, "readWAV: warning - more tags (%d extra bytes)\n",
		filesize - block.size - 8);
    if(block.size + 8 > filesize)
	fprintf(stderr, "readWAV: warning - short file (%d bytes missing)\n",
		block.size + 8 -  filesize);
    if(reader_read(b, 1, 4, fi) < 4) {
	return 0;
    }
    if(strncmp((char*)b, "WAVE", 4)) {
	fprintf(stderr, "readWAV: not a WAV file (2)\n");
	return 0;
    }
 
    do
    {
	if(!getWAVBlock(fi, &block)) {
	    break;
	}

	pos = reader_getpos(fi);
	if(!strncmp(block.id, "fmt ", 4)) {
	    if(reader_read(&b, 1, 16, fi)<16) {
		fprintf(stderr, "readWAV: Couldn't get fmt block\n");
		return 0;
	    }
	    wav->tag = b[0]|b[1]<<8;
	    wav->channels = b[2]|b[3]<<8;
	    wav->sampsPerSec = b[4]|b[5]<<8|b[6]<<16|b[7]<<24;
	    wav->bytesPerSec = b[8]|b[9]<<8|b[10]<<16|b[11]<<24;
	    wav->align = b[12]|b[13]<<8;
	    wav->bps = b[14]|b[15]<<8;
	} else if (!strncmp(block.id, "LIST", 4)) {
	    // subchunk ICMT (comment) may exist
	} else if (!strncmp(block.id, "data", 4)) {
	    wav->data = (unsigned char*)malloc(block.size);
	    if(!wav->data) {
		fprintf(stderr, "Out of memory (%d bytes needed)", block.size);
		return 0;
	    }
	    if(reader_read(wav->data, 1, block.size, fi) < block.size) {
		fprintf(stderr, "readWAV: Couldn't get data block\n");
		return 0;
	    }
	    wav->size = block.size;
	}
	pos+=block.size;
	reader_goto(fi, pos);
    }
    while (pos < filesize);
    if(!wav->data)
	return 0;

    return 1;
}

int writeWAV(char*filename, struct WAV*wav)
{
    FILE*fi = fopen(filename, "wb");
    char*b="RIFFWAVEfmt \x10\0\0\0data";
    char c[16];
    unsigned long int w32;
    if(!fi)
	return 0;
    fwrite(b, 4, 1, fi);
    w32=(/*fmt*/8+0x10+/*data*/8+wav->size);
    c[0] = w32;
    c[1] = w32>>8;
    c[2] = w32>>16;
    c[3] = w32>>24;
    fwrite(c, 4, 1, fi);
    fwrite(&b[4], 12, 1, fi);
    c[0] = wav->tag;
    c[1] = wav->tag>>8;
    c[2] = wav->channels;
    c[3] = wav->channels>>8;
    c[4] = wav->sampsPerSec;
    c[5] = wav->sampsPerSec>>8;
    c[6] = wav->sampsPerSec>>16;
    c[7] = wav->sampsPerSec>>24;
    c[8] = wav->bytesPerSec;
    c[9] = wav->bytesPerSec>>8;
    c[10] = wav->bytesPerSec>>16;
    c[11] = wav->bytesPerSec>>24;
    c[12] = wav->align;
    c[13] = wav->align>>8;
    c[14] = wav->bps;
    c[15] = wav->bps>>8;
    fwrite(c, 16, 1, fi);
    fwrite(&b[16], 4, 1, fi);
    c[0] = wav->size;
    c[1] = wav->size>>8;
    c[2] = wav->size>>16;
    c[3] = wav->size>>24;
    fwrite(c,4,1,fi);
    fwrite(wav->data,wav->size,1,fi);
    fclose(fi);
    return 1;
}

void printWAVInfo(struct WAV*wav)
{
    printf("tag:%04x channels:%d samples/sec:%d bytes/sec:%d align:%d bits/sample:%d size:%d\n",
	    wav->tag, wav->channels, wav->sampsPerSec, wav->bytesPerSec, 
	    wav->align, wav->bps, wav->size);
}

int convertWAV2mono(struct WAV*src, struct WAV*dest, int rate)
{
    int samplelen=src->size/src->align;
    int bps=src->bps;
    double ratio;
    double pos = 0;
    int pos2 = 0;
    int channels=src->channels;
    int i;
    int fill;

    dest->sampsPerSec = rate;
    dest->bps = 16;
    dest->channels = 1;
    dest->align = 2;
    dest->tag = src->tag;
    dest->bytesPerSec = dest->sampsPerSec*dest->align;

    ratio = (double)dest->sampsPerSec/(double)src->sampsPerSec;
    fill = (int)(ratio+1)*2;
    
    dest->data = (unsigned char*)malloc((int)(samplelen*ratio*2)+128);
    if(!dest->data) 
	return 0;
    dest->size = (int)(samplelen*ratio)*2;

    if(bps == 8) {
	if(ratio <= 1) {
	    for(i=0; i<src->size; i+=channels) {
		int pos2 = ((int)pos)*2;
		int data = (src->data[i]+src->data[i+channels-1])/2;
		dest->data[pos2+0] = 0;
		dest->data[pos2+1] = data + 128;
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		int data = (src->data[i]+src->data[i+channels-1])/2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = 0;
		    dest->data[pos2+j+1] = data + 128;
		}
		pos += ratio;
	    }
	}
    } else if(bps == 16) {
	if(ratio <= 1) {
	    for(i=0; i<src->size/2; i+=channels) {
		int pos2 = ((int)pos)*2;
		dest->data[pos2+0]=src->data[i*2];
		dest->data[pos2+1]=src->data[i*2+1];
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size/2; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = src->data[i*2];
		    dest->data[pos2+j+1] = src->data[i*2+1];
		}
		pos += ratio;
	    }
	}
    } else if(bps == 32) {
	if(ratio <= 1) {
	    for(i=0; i<src->size/4; i+=channels) {
		int pos2 = ((int)pos)*2;
		dest->data[pos2+0]=src->data[i*4+2];
		dest->data[pos2+1]=src->data[i*4+3];
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size/4; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = src->data[i*4+2];
		    dest->data[pos2+j+1] = src->data[i*4+3];
		}
		pos += ratio;
	    }
	}
    }
    return 1;
}

unsigned char table_lawmix[256][256];
unsigned char table_wav2law[32768];
unsigned char imap[256];
short int table_law2wav[256];
char tables = 0; //set to 1 once tables have been generated
void make_tables();

unsigned char inline wavsample2law_fast(short int val)
{
    if(val>0) {
	return table_wav2law[val]|1;
    } else {
	return table_wav2law[-val];
    }
}

short int inline lawsample2wav_fast(unsigned char val)
{
    return table_law2wav[val];
}

/* only used in make_tables*/
unsigned char wavsample2law(short int val) 
{
    char sign;
    unsigned char carry;
    unsigned char alaw;

    sign=val>=0?1:0;
    if(val<0)
	val*=-1;

    if (val < 0x0800) {
        if (val < 0x0200) {
            if (val < 0x0100) {
		carry = (val&12)?1:0;
                val&=0xF0; val>>=3;
		alaw=0x00+val+carry;
            } else {
		carry = (val&28)?1:0;
                val&=0xE0; val>>=4;
		alaw=0x20+val+carry;
            }
        } else {
            if (val < 0x0400)
            {
		carry = (val&60)?1:0;
                val&=0x1C0; val>>=5;
		alaw=0x30+val+carry;
            } else {
		carry = (val&124)?1:0;
                val&=0x380; val>>=6;
		alaw=0x40+val+carry;
            }
        }
    } else {
        if (val < 0x2000) {
            if (val < 0x1000) {
		carry = (val&252)?1:0;
                val&=0x700; val>>=7;
		alaw=0x50+val+carry;
            } else {
		carry = (val&508)?1:0;
                val&=0xE00; val>>=8;
		alaw=0x60+val+carry;

            }
        } else {
            if (val < 0x4000)
            {
		carry = (val&1020)?1:0;
                val&=0x1C00; val>>=9;
		alaw=0x70+val+carry;

            } else {
		alaw = 0x7f;
	    }
        }
    }
    alaw += sign*0x80;
    return imap[(alaw ^ 0x55)];
}

/* only used in make_tables*/
short int lawsample2wav(unsigned char value)
{
    unsigned char segment;
    unsigned char offset;
    short int pcm;
    value = imap[value];
    value^=0x55;
    segment = value & 0x70;
    segment >>= 4;
    offset=value & 0x0f;
    if (segment < 0x04) {
        if (segment < 0x02) {
            if (segment < 0x01)
                pcm=(0x0002+(offset<<3));
            else
                pcm=(0x0082+(offset<<3));
        } else {
            if (segment < 0x03)
                pcm=(0x0102+(offset<<4));
            else
                pcm=(0x0202+(offset<<5));
        }
    } else {
        if (segment < 0x06) {
            if (segment < 0x05)
                pcm=(0x0402+(offset<<6));
            else
                pcm=(0x0802+(offset<<7));
        } else {
            if (segment < 0x07)
                pcm=(0x1002+(offset<<8));
            else
                pcm=(0x2002+(offset<<9));
        }
    }
    if (!(value&0x80))
        pcm*=-1;
    return pcm;
}


void convertWAV2LAW(struct WAV*src, unsigned char**data, int*size)
{
    struct WAV mono;
    int t;
    if(!tables)
	make_tables();
    convertWAV2mono(src, &mono, 8000);
    *size = mono.size/2;
    *data = (unsigned char*)malloc(*size);

    for(t=0;t<mono.size/2;t++) {
	signed short value = ((signed short*)mono.data)[t];
	(*data)[t]=wavsample2law_fast(value);
    }
    free(mono.data);
}

void convertLAW2WAV(void*data, int size, struct WAV*dest)
{
    unsigned char*d = (unsigned char*)data;
    int t;
    if(!tables)
	make_tables();

    dest->data = (unsigned char*)malloc(size*2);
    dest->size = size*2;
    dest->sampsPerSec = 8000;
    dest->bps = 16;
    dest->channels = 1;
    dest->align = 2;
    dest->tag = 0x0001;
    dest->bytesPerSec = dest->sampsPerSec*dest->align;

    for(t=0;t<size;t++) {
	((signed short*)dest->data)[t] = lawsample2wav_fast(*d);
	d++;
    }
}

int readWAV(char* filename, struct WAV*wav)
{
    struct reader r;
    FILE*fi = fopen(filename, "rb");
    if(!fi) {
	fprintf(stderr, "Couldn't open %s\n", filename);
	return 0;
    }
    r.fi = fi;
    r.data = 0;
    r.len = 0;
    r.pos = 0;
    return readWAV2(&r,wav);
}

int readWAVFromMem(void*data, int len, struct WAV*wav)
{
    struct reader r;
    r.fi = 0;
    r.data = (unsigned char*)data;
    r.len = len;
    r.pos = 0;
    return readWAV2(&r,wav);
}

void make_tables()
{
    int s,t,u;

    /* cute little bit reversal routine 
       (1 -> 0x80, 2->0x40, 3->0xc0 etc.) */
    for(t=0,s=0,u=0;t<(imap[t]=s,u=256);t++) 
	do u>>=1,s^=u; while((s^u)>s);

    for(t=0;t<32768;t++) {
	table_wav2law[t] = wavsample2law(t)&(~1);
    }
    for(t=0;t<256;t++) {
	table_law2wav[t] = lawsample2wav(t);
    }

    for(s=0;s<256;s++) {
	for(t=0;t<256;t++) {
	    int sample1 = lawsample2wav_fast(s);
	    int sample2 = lawsample2wav_fast(t);
	    int sample = sample1+sample2;
	    if(sample>32767) sample=32767;
	    if(sample<-32768) sample=-32768;
	    table_lawmix[s][t] = wavsample2law_fast(sample);
	}
    }
    tables = 1;
}

unsigned char*wav2law(unsigned char*wavedata, int len, int *len2)
{
    struct WAV wav;
    unsigned char* data;
    int size;
    if(!readWAVFromMem(wavedata, len, &wav))
	return 0;
    convertWAV2LAW(&wav, &data, &size);
    free(wav.data);
    *len2 = size;
    return data;
}

void mix_law(unsigned char*block1, unsigned char*block2, unsigned char*block3, int len)
{
    int t;
    for(t=0;t<len;t++) {
	block3[t] = table_lawmix[block1[t]][block2[t]];
    }
}

